# -*- coding: utf-8 -*-
"""
Created on Mon Oct  1 20:04:24 2018


@author: rday

Various functions relevant to rotations

"""
import numpy as np
from math import atan2

    
def Euler(rotation):
    '''
    Euler rotation angle generation, Z-Y-Z convention, as defined for a user-defined
    rotation matrix. Special case for B = +/- Z*pi where conventional approach doesn't
    work due to division by zero, then Euler_A is zero and Euler_y is arctan(R10,R00)

    
    *args*:

        - **rotation**: numpy array of 3x3 float (rotation matrix)
        OR tuple/list of vector and angle (numpy array of 3 float, float) respectively
    
    *return*:

        - **Euler_A**, **Euler_B**, **Euler_y**: float, Euler angles associated with
        the given rotation.
        
    ***
    '''
    
    if type(rotation)==np.ndarray:
        rot_mat = rotation
    else:
        
        rot_mat = Rodrigues_Rmat(rotation[0],rotation[1])
    if abs(rot_mat[2,2])!=1.0:
    
    	Euler_A = atan2(rot_mat[1,2],rot_mat[0,2])
    
    	Euler_y = atan2(rot_mat[2,1],-rot_mat[2,0])
    
    	Euler_B = np.arccos(rot_mat[2,2])
    	if Euler_B>np.pi:
    		Euler_B = 2*np.pi- Euler_B
    
    	
    else:
        if np.sign(rot_mat[2,2])==1.0:
            Euler_B = 0.0
        else:
            Euler_B = np.pi
        if np.sign(rot_mat[2,2])==1:
            Euler_A = 0
            Euler_y = atan2(rot_mat[1,0],rot_mat[0,0])
        else:
            Euler_A = 0
            Euler_y = atan2(rot_mat[1,0],-rot_mat[0,0])

    return Euler_A,Euler_B,Euler_y



def Euler_to_R(Euler_A,Euler_B,Euler_y):
    
    '''
    Inverse of *Euler*, generate a rotation matrix from the Euler angles A,B,y
    with the same conventiona as in *Euler*.
   
    *args*:

        - **Euler_A**, **Euler_B**, **Euler_y**: float
    
    *return*:

        - numpy array of 3x3 float
        
    ***
    '''
    
    rot_y = Rodrigues_Rmat(np.array([0,0,1]),Euler_y)
    rot_B = Rodrigues_Rmat(np.array([0,1,0]),Euler_B)
    rot_A = Rodrigues_Rmat(np.array([0,0,1]),Euler_A)
    
    return np.dot(rot_A,np.dot(rot_B,rot_y))


    

def rotate_v1v2(v1,v2):
    
    '''
    This generates the rotation matrix for PRE-multiplication rotation:
    or written another way, defining R s.t. R.v1 = v2. This rotation will rotate
    the vector **v1** onto the vector **v2**.
    
    *args*:

        - **v1**: numpy array len 3 of float, input vector
        
        - **v2**: numpy array len 3 of float, vector to rotate into
        
    *return*:

        - **Rmat**: numpy array of 3x3 float, rotation matrix
    
    ***    
    '''
    
    v1 = v1/np.linalg.norm(v1)
    v2 = v2/np.linalg.norm(v2)
    vx=np.cross(v1,v2)
    cos=np.dot(v1,v2)
    vm=np.array([[0.0,-vx[2],vx[1]],[vx[2],0.0,-vx[0]],[-vx[1],vx[0],0.0]])
    if abs(cos+1)>1e-10:
        Rmat=np.identity(3)+vm+np.dot(vm,vm)/(1+cos)
    else:
        Rmat = np.identity(3)
        Rmat[2,2] = -1.0
    return Rmat


def Rodrigues_Rmat(nvec,theta):
    '''
    Following Rodrigues theorem for rotations, define a rotation matrix which
    corresponds to the rotation about a vector nvec by the angle theta, in radians.
    Works in pre-multiplication order (i.e. v' = R.v)
    
    *args*:

        - **nvec**: numpy array len 3 axis of rotation
        
        - **theta**: float radian angle of rotation counter clockwise for theta>0
        
    *return*:

        - **Rmat**: numpy array 3x3 of float rotation matrix
        
    ***    
    '''
    
    nvec = nvec/np.linalg.norm(nvec)
    Kmat = np.array([[0,-nvec[2],nvec[1]],[nvec[2],0,-nvec[0]],[-nvec[1],nvec[0],0]])
    Rmat = np.identity(3)+np.sin(theta)*Kmat+(1-np.cos(theta))*np.dot(Kmat,Kmat)
    return Rmat   


def rot_vector(Rmatrix):
    
    '''
    Inverse to *Rodrigues_Rmat*, take rotation matrix as input and return 
    the angle-axis convention rotations corresponding to this rotation matrix.
    
    *args*:

        - **Rmatrix**: numpy array of 3x3 float, rotation matrix
        
    *return*:
    
        - **nvec**: numpy array of 3 float, rotation axis
        
        - **theta**: float, rotation angle in float
        
    ***
    '''
    L,nvec=np.linalg.eig(Rmatrix)
    nvec = np.real(nvec[:,np.where(abs(L-1)<1e-10)[0][0]])
    theta = np.arccos((np.trace(Rmatrix)-1)/2)
    R_tmp = Rodrigues_Rmat(nvec,theta)
    if np.linalg.norm(R_tmp-Rmatrix)<1e-10:
        return nvec,theta
    else:
        R_tmp = Rodrigues_Rmat(nvec,-theta)
        if np.linalg.norm(R_tmp-Rmatrix)<1e-10:
            return nvec,-theta
        else:
            print('ERROR: COULD NOT DEFINE ROTATION MATRIX FOR SUGGESTED BASIS TRANSFORMATION!')
            return None


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#Created on Mon Mar 18 13:27:50 2019


#@author: ryanday
#MIT License

#Copyright (c) 2018 Ryan Patrick Day

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.


import numpy as np
import datetime as dt

class intensity_map:

    '''
    Class for organization and saving of data, as well as metadata related to
    a specific ARPES calculation.
    '''
    
    def __init__(self,index,Imat,cube,kz,T,hv,pol,dE,dk,self_energy=None,spin=None,rot=0.0,notes=None):
        self.index = index
        self.cube = cube
        self.kz = kz
        self.T = T
        self.hv = hv
        self.pol = pol
        self.spin = spin
        self.Imat = Imat
        self.rot = rot
        self.self_energy = self_energy
        self.dE = dE
        self.dk = dk
        if type(notes)==str:
            self.notes = notes
        else:
            self.notes = 'N/A'
        
    def save_map(self,directory):

        '''

        Save the intensity map: if 2D, just a single file, if 3D, each constant-energy
        slice is saved separately. Saved as .txt file

        *args*:

            - **directory**: string, directory for saving intensity map to
        
        *return*:

            - boolean

        ***
        '''
        self.write_meta(directory+'_meta.txt')
        dim_min = min([a for a in np.shape(self.Imat)])
        if len(np.shape(self.Imat))==3 and dim_min>1:
            for i in range(np.shape(self.Imat)[2]):
                filename = directory + '_{:d}.txt'.format(i)
                self.write_2D_Imat(filename,i)
        else:
            self.write_2D_Imat(directory+'_0.txt',-1)
    
        return True
    
    def write_2D_Imat(self,filename,index):
        
        '''
        Sub-function for producing the textfiles associated with a 2dimensional numpy array of float

        *args*:

            - **filename**: string, indicating destination of file
            
            - **index**: int, energy index of map to save, if -1, then just a 2D map, and save the whole
            thing
            
        *return*:

            - boolean
        
        ***
        '''
        if index>-1:
            imat = self.Imat[:,:,index]
        else:
            imat = np.squeeze(self.Imat)
                
        with open(filename,"w") as destination:        
            
            for ii in range(np.shape(imat)[0]):
                tmpline = " ".join(['{:0.04f}'.format(vi) for vi in imat[ii,:]])
                tmpline+="\n"
                destination.write(tmpline)
        destination.close()
        return True
    
    def write_meta(self,destination):
        
        '''
        Write meta-data file for ARPES intensity calculation.
        
        *args*:

            - **destination**: string, file-lead
            
        *return*:

            - boolean

        ***
        '''
        
        with open(destination,'w') as tofile:
            now = dt.datetime.now()
            now_time = now.strftime('%H:%M:%S %d/%m/%y')
            tofile.write('ARPES calculation: '+now_time+'\n')
            tofile.write('Calculation notes: {:s}\n\n'.format(self.notes))
            tofile.write('Temperature: {:0.02f} K\n'.format(self.T))
            tofile.write('Photon Energy: {:0.04f} eV\n'.format(self.hv))
            tofile.write('Polarization: [{:s}, {:s}, {:s}]\n'.format(*self.pol.astype(str)))
            tofile.write('Energy Resolution: {:0.04f} eV\n'.format(self.dE))
            tofile.write('Momentum Resolution: {:0.04f} 1/A\n\n'.format(self.dk))
            tofile.write('Kx Domain: {:0.04f} -> {:0.04f} 1/A, N = {:d}\n'.format(*self.cube[0]))
            tofile.write('Ky Domain: {:0.04f} -> {:0.04f} 1/A, N = {:d}\n'.format(*self.cube[1]))
            tofile.write('Kz: {:0.04f} 1/A\n'.format(self.kz))
            tofile.write('Energy Domain: {:0.04f} -> {:0.04f} eV, N = {:d}\n\n'.format(*self.cube[2]))
            tofile.write('Sample Rotation: {:0.04f}\n'.format(self.rot))
            if self.spin is None:
                tofile.write('Spin Projection: None\n')
                tofile.write('Spin Axis: None \n')
            else:
                tofile.write('Spin Projection: {:d}\n'.format(self.spin[0]))
                tofile.write('Spin Axis: {:0.04f}, {:0.04f}, {:0.04f}\n'.format(*self.spin[1:]))
            
            if self.self_energy[0]=='poly':
                tofile.write('Self Energy: '+'+'.join(['{:0.04f}w^{:d}'.format(self.self_energy[i],i-1) for i in range(1,len(self.self_energy))])+'\n')
            elif self.self_energy[0] == 'constant':
                tofile.write('Self Energy: {:0.04f}\n'.format(self.self_energy[1]))
            else:
                if self.self_energy[0] == 'func':
                    w = np.linspace(*self.cube[2])
                    SE = self.self_energy[1](w)
                else:
                    w = self.self_energy[1]
                    SE = self.self_energy[2]
                tofile.write('Energy (eV) | Self Energy (eV) \n')
                for ii in range(w):
                    tofile.write('{:s}\t {:s}\n'.format(str(w[ii]),str(SE[ii])))
            
        tofile.close()
        
        
    def copy(self):

        '''
        Copy-by-value of the intensity map object. 

        *return*:

            - *intensity_map* object with identical attributes to self.

        ***
        '''
        return intensity_map(self.index,self.Imat,self.cube,self.kz,self.T,self.hv,self.pol,self.dE,self.dk,self.self_energy,self.spin,self.rot)



# -*- coding: utf-8 -*-

#Created on Mon Aug 14 09:32:57 2017

#@author: ryanday
#MIT License

#Copyright (c) 2018 Ryan Patrick Day

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

import linecache
import pkg_resources



a_file = 'atomic_mass.txt'
filename = pkg_resources.resource_filename(__name__,a_file)
def get_mass_from_number(N_at):
    '''

    Pull atomic mass for the indicated atomic number
    
    *args*:

        - **N_at**: int, atomic number
        
    *return*:

        - float, atomic mass, in atomic mass units
    
    ***
    '''
    try:
        return float(linecache.getline(filename,int(N_at)).split('\t')[2][:-1])
    except IndexError:
        print('ERROR: Invalid atomic number, returning mass = 0.')
        return 0.0

def get_el_from_number(N_at):
    
    '''
    Get symbol for element, given the atomic number
    
    *args*:

        - **N_at**: int, atomic number
    
    *return*:

        - string, symbol for element
    
    ***
    '''
    
    try:
        return linecache.getline(filename,int(N_at)).split('\t')[1]
    except IndexError:
        print('ERROR: Invalid atomic number, returning empty string.')
        return ''

def get_num_from_el(el):
    
    '''

    Get atomic number from the symbol for the associated element. Returns 0 for
    invalid entry.
    
    *args*:

        - **el**: string, symbol for element    
    
    *return*:

        - **Z**: int, atomic number.
    
    '''
    Z  = -1
    with open(filename,'r') as mass:
        for l in mass:
            line = l.split('\t')
            if line[1]==el:
                Z = int(line[0])
    mass.close()
    if Z == -1:
        print('WARNING!! Invalid symbol passed. Returning with Z = 0')
        Z = 0
        
    return Z


# -*- coding: utf-8 -*-

#Created on Tue Jan 29 16:02:40 2019

#@author: ryanday
#MIT License

#Copyright (c) 2018 Ryan Patrick Day

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

import numpy as np
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt
import chinook.adaptive_int as adint
import chinook.electron_configs as econ

####PHYSICAL CONSTANTS RELEVANT TO CALCULATION#######
hb = 6.626*10**-34/(2*np.pi)
c  = 3.0*10**8
q = 1.602*10**-19
A = 10.0**-10
me = 9.11*10**-31
mN = 1.67*10**-27
kb = 1.38*10**-23


def make_radint_pointer(rad_dict,basis,Eb):
    
    '''
    Define executable radial integral functions, and store in a 
    pointer-integer referenced array. This allows for fewer executions
    of the interpolation function in the event where several orbitals
    in the basis share the same a,n,l. Each of these gets 2 functions
    for l +/-1, which are stored in the rows of the array **B_array**.
    The orbitals in the basis then are matched to these executables,
    with the corresponding executable row index saved in **B_pointers**.
    
    Begin by defining the executable radial wavefunctions, then perform
    integration at several binding energies, finally returning an
    interpolation of these integrations.
    
    *args*: 

        - **rad_dict**: dictionary of ARPES parameters: relevant keys are 
        'hv' (photon energy), 'W' (work function), and the rad_type
        (radial wavefunction type, as well as any relevant additional
        pars, c.f. *radint_lib.define_radial_wavefunctions*).
        Note: *'rad_type'* is optional, (as is *rad_args*, depending on choice
        of radial wavefunction.)
        
        - **basis**: list of orbitals in the basis
        
        - **Eb**: tuple of 2 floats indicating the range of energy of
        interest (increasing order)
        
    *return*: 

        - **B_array**: numpy array of Nx2 executable functions of float
        
        - **B_pointers**: numpy array of integer indices matching orbital
        basis ordering to the functions in **B_array**
        
    ***   
    '''
    radial_funcs = define_radial_wavefunctions(rad_dict,basis)
    fixed = True if ('rad_type' in rad_dict.keys() and rad_dict['rad_type']=='fixed') else False
    B_dictionary = fill_radint_dic(Eb,radial_funcs,rad_dict['hv'],rad_dict['W'],rad_dict['phase_shifts'],fixed)
    B_array,B_pointers = radint_dict_to_arr(B_dictionary,basis)
    
    return B_array,B_pointers


def radint_dict_to_arr(Bdict,basis):

    '''
    Take a dictionary of executables defined for different combinations
    of a,n,l and send them to an array, with a corresponding pointer
    array which can be used to dereference the relevant executable.
    
    *args*:     

        - **Bdict**: dictionary of executables with 'a-n-l-l'' keys
        
        - **basis**: list of orbital objects
    
    *return*:  

        - **Blist**: numpy array of the executables, organized by a-n-l,
        and l' (size Nx2, where N is the length of the set of 
        distinct a-n-l triplets)
        
        - **pointers**: numpy array of length (basis), integer datatype
        indicating the related positions in the **Blist** array
    
    ***
    '''
    Blist = []
    lchan = []
    pointers = []
    for b in basis:
        orbstr = '{:d}-{:d}-{:d}'.format(b.atom,b.n,b.l)
        if orbstr not in lchan:
            lchan.append(orbstr)
            Blist.append([Bdict[orbstr+'-{:d}'.format(b.l-1)],Bdict[orbstr+'-{:d}'.format(b.l+1)]])
        pointers.append(lchan.index(orbstr))
    return np.array(Blist),np.array(pointers)
        


def define_radial_wavefunctions(rad_dict,basis):
    
    '''
    Define the executable radial wavefunctions for computation of
    the radial integrals
    
    *args*:   

        - **rad_dict**: essential key is *'rad_type'*, if not passed,
        assume Slater orbitals. 
            
        - **rad_dict['rad_type']**:

                - *'slater'*: default value, if *'rad_type'* is not passed,
                Slater type orbitals assumed and evaluated for the integral
                    - *'rad_args'*: dictionary of float, supplying optional final-state 
                    phase shifts, accounting for scattering-type final states. keys of form
                    'a-n-l-lp'. Radial integrals will be accordingly multiplied
                
                - *'hydrogenic'*: similar in execution to *'slater'*, 
                but uses Hydrogenic orbitals--more realistic for light-atoms
                    - *'rad_args'*: dictionary of float, supplying optional final-state 
                    phase shifts, accounting for scattering-type final states. keys of form
                    'a-n-l-lp'. Radial integrals will be accordingly multiplied
                
                - *'grid'*: radial wavefunctions evaluated on a grid of
                radii. Requires also another key_value pair:
                    
                   - *'rad_args'*: dictionary of numpy arrays evaluating
                   the radial wavefunctions. Requires an *'r'* array,
                   as well as 'a-n-l' indicating 'atom-principal quantum number-orbital angular momentum'.
                   Must pass such a grid for each orbital in the basis!
                   
                - *'exec'*: executable functions for each 'a-n-l' i.e.
                'atom-principal quantum number-orbital angular momentum'. 
                If 	executable is chosen, require also:
                    
                        - *'rad_args'*, which will be a dictionary of
                        executables, labelled by the keys 'a-n-l'. 
                        These will be passed to the integral routine.
                        Note that it is required that the executables
                        are localized, i.e. vanishing for large radii.
                        
                - *'fixed*: radial integrals taken to be constant float,
                require dictionary:
                    
                    - *'rad_args'* with keys 'a-n-l-lp', i.e. 
                    'atom-principal quantum number-orbital angular momentum-final state angular momentum'
                    and complex float values for the radial integrals.
                    
        - **basis**: list of orbital objects
        
    *return*: 

        **orbital_funcs**: dictionary of executables
        
    ***    
    '''
    
    orbital_funcs = {}
    orbitals = gen_orb_labels(basis)

    if 'rad_type' not in rad_dict.keys():
        rad_dict['rad_type'] = 'slater'
    
    if rad_dict['rad_type'].lower()=='slater':
        for o in orbitals:
            orbital_funcs[o] = econ.Slater_exec(*orbitals[o])
    
    elif rad_dict['rad_type'].lower() == 'hydrogenic':
        for o in orbitals:
            orbital_funcs[o] = econ.hydrogenic_exec(*orbitals[o])
        
    elif rad_dict['rad_type'].lower() == 'grid':
        if 'rad_args' not in rad_dict.keys():
            print('ERROR: No "rad_args" key passing the grid values passed to ARPES calculation.\n Exiting. \n See Radial Integrals in the Manual for further details.\n')
            return None
        elif np.sum([o in rad_dict['rad_args'].keys() for o in orbitals])<len(orbitals):
            print('ERROR: Missing radial wavefunction grids--confirm all atoms and orbital shells have a grid.\n Exiting.\n')
            return None
        elif 'r' not in rad_dict['rad_args'].keys():
            print('ERROR: no radial grid passed for proper scaling of the radial wavefunctions.\n Exiting.\n')
            return None
        if rad_dict['rad_args']['r'].min()>0:
            print('ERROR: radial grid must go to the origin.\n Exiting.\n')
            return None
        else:
            for o in orbitals:
                orbital_funcs[o] = interp1d(rad_dict['rad_args']['r'],rad_dict['rad_args'][o])
    
    elif rad_dict['rad_type'].lower() == 'exec':
        if 'rad_args' not in rad_dict.keys():
            print('ERROR: No "rad_args" key passing an executable to ARPES calculation.\n. Exiting.\n See Radial Integrals in the documentation for further details.\n')
            return None
        elif np.sum([o in rad_dict['rad_args'].keys() for o in orbitals])<len(orbitals):
            print('ERROR: Missing radial wavefunction functionss--confirm all atoms and orbital shells have a function.\n Exiting.\n')
            return None
        else:
            for o in orbitals:
                orbital_funcs[o] = rad_dict['rad_args'][o]
    
    
    elif rad_dict['rad_type'].lower() == 'fixed':    
        if 'rad_args' not in rad_dict.keys():
            print('ERROR: Missing radial integral values.\n Exiting.\n See Radial Integrals in the documentation for further details.\n')
            return None
        for o in basis:
            lp = np.array([o.l-1,o.l+1])
            for lpi in lp:
                ostr = '{:d}-{:d}-{:d}-{:d}'.format(o.atom,o.n,o.l,lpi)
                if lpi>=0:
                    orbital_funcs[ostr] = gen_const(rad_dict['rad_args'][ostr])
                else:
                    orbital_funcs[ostr] = gen_const(0.0)
    
    return orbital_funcs
        
    
    
def gen_orb_labels(basis):
    
    '''
    Simple utility function for generating a dictionary of 
    atom-n-l:[Z, orbital label] pairs, to establish which radial integrals
    need be computed.
    
    *args*:   

        - **basis**: list of orbitals in basis
    
    
    *return*:

        - **orbitals**: dictionary of radial integral pairs
        
    ***    
    '''
    orbitals = {}
    for o in basis:
        o_str = '{:d}-{:d}-{:d}'.format(o.atom,o.n,o.l)
        if o_str not in orbitals.keys():
            orbitals[o_str] = [o.Z,o.label]
    return orbitals
    

def radint_calc(k_norm,orbital_funcs,phase_shifts=None):
    
    '''
    Compute dictionary of radial integrals evaluated at a single |k| value
    for the whole basis. Will avoid redundant integrations by checking for
    the presence of an identical dictionary key. The integration is done
    as a simple adaptive integration algorithm, defined in the 
    *adaptive_int* library.
    
    *args*: 

        - **k_norm**: float, length of the k-vector 
        (as an argument for the spherical Bessel Function)
        
        - **orbital_funcs**: dictionary, radial wavefunction executables
        
    *kwargs*:

        - **phase_shifts**: dictionary of phase shifts, to convey final state scattering
        
    *returns*:

        - **Bdic**: dictionary, key value pairs in form -- 'ATOM-N-L':*Bval*
        
    ***
    '''
    
    Bdic = {}
    
    for o in orbital_funcs:
        
            
        l = int(o.split('-')[2])
        L=[x for x in [l-1,l+1]]
        for lp in L:
            Blabel=o+'-'+str(lp)
            if phase_shifts is None:
                phase_factor = 1.0
            else:
                phase_factor = phase_shifts[Blabel]
            if Blabel not in Bdic.keys():
                if lp<0:
                    tmp_B = 0.0
                else:
                
                    integrand = adint.general_Bnl_integrand(orbital_funcs[o],k_norm,lp)
                    b = find_cutoff(integrand)
                    a = 0.001
                        
                    tol = 10.0**-10
                        
                    tmp_B = phase_factor*adint.integrate(integrand,a,b,tol)
                       
                Bdic[Blabel]=tmp_B
    return Bdic

def fill_radint_dic(Eb,orbital_funcs,hv,W=0.0,phase_shifts=None,fixed=False):
    
    '''
    Function for computing dictionary of radial integrals. 
    Can pass either an array of binding energies or a single binding
    energy as a float. In either case, returns a dictionary however 
    the difference being that the key value pairs will have a value
    which is itself either a float, or an interpolation mesh over 
    the range of the binding energy array. The output can then be 
    used by either writing **Bdic['key']** or
    **Bdic['key']**(valid float between endpoints of input array)
    
    *args*:   

        - **Eb**: float or tuple indicating the extremal energies
       
        - **orbital_funcs**: dictionary of executable orbital radial wavefunctions
       
        - **fixed**: bool, if True, constant radial integral for each scattering
        channel available: then the orbital_funcs dictionary already
        has the radial integral evaluated
      
        - **hv**: float, photon energy of incident light.
        
    *kwargs*:

        - **W**: float, work function
        
        - **phase_shifts**: dictionary for final state phase shifts, as an optional
        extension beyond pure- free electron final states. For now, float type.
        
    *return*:

        - **Brad**: dictionary of executable interpolation grids

    ***
    '''
    if fixed:
        Brad = {o:orbital_funcs[o] for o in orbital_funcs}
    else:

        if type(Eb)==float:
            kval = np.sqrt(2.0*me/hb**2*((hv-W)+Eb)*q)*A 
            return (radint_calc(kval,orbital_funcs) if ((hv-W)+Eb)>=0 else 0)
        elif type(Eb)==tuple:
            Brad_es=np.linspace(Eb[0],Eb[-1],5)
            BD_coarse={}
            for en in Brad_es:
                k_coarse = np.sqrt(2.0*me/hb**2*((hv-W)+en)*q)*A #calculate full 3-D k vector at this surface k-point given the incident radiation wavelength, and the energy eigenvalue, note binding energy follows opposite sign convention
                tmp_Bdic = (radint_calc(k_coarse,orbital_funcs,phase_shifts) if ((hv-W)+en)>=0 else {})
                for b in tmp_Bdic:
                    try:
                        BD_coarse[b].append(tmp_Bdic[b])
                    except KeyError:
                        BD_coarse[b] = [tmp_Bdic[b]]
    
            Brad = {}     
            for b in BD_coarse:
                f = interp1d(Brad_es,BD_coarse[b],kind='cubic')
                Brad[b] = f
    if type(Eb)!=float and type(Eb)!=tuple and type(Eb)!=list and type(Eb)!=np.ndarray:
        print('RADIAL INTEGRAL ERROR: Enter a valid energy (float) or range of energies (tuple of 2 floats indicating endpoints). Returning None')
        return None
    else:
        return Brad

def find_cutoff(func):
    
    '''
    
    Find a suitable cutoff lengthscale for the radial integration:
    Evaluate the function over a range of 20 Angstrom, with reasonable
    detail (dr = 0.02 A). Find the maximum in this range. The cutoff
    tolerance is set to 1/1e4 of the maximum value. Since this 'max'
    is actually a lower bound on the true maximum, this will only give
    us a more strict cutoff tolerance than is absolutely possible. With
    this point found, we then find all points which are within the
    tolerance of zero. The frequency of these points is then found. When
    the frequency is constant and 1 for all subsequent points, we have
    found the point of convergence. If the 'point of convergence' is the
    last point in the array, the radial wavefunction really isn't suitably
    localized and the user should not proceed without giving more
    consideration to the application of the LCAO approximation to such
    a function.
    
    *args*: 

        - **func**: the integrand executable
    
    *return*:

        - float, cutoff distance for integration
        
    ***    
    '''
    
    r_c = np.linspace(0,25,1000)[1:]
    func_c = func(r_c)
    f_max = abs(func_c).max()

    ftol = f_max/1e4
    sub_tol_inds = np.where(abs(func_c)<ftol)[0]
    d_inds = [sub_tol_inds[i+1]-sub_tol_inds[i] for i in range(len(sub_tol_inds[:-1]))]
    for i in range(len(d_inds)):
        if np.sum(d_inds[i:]) == len(d_inds[i:]):
            cutoff = i
            break
#
    if cutoff==len(d_inds):
        print('RADIAL WAVEFUNCTION ERROR: WAVE FUNCTION IS NOT WELL CONVERGED OVER 25 ANGSTROM. EXITING')
        f_max_ind = np.where(abs(func_c)==f_max)[0][0]
        sub_tol_pts = r_c[sub_tol_inds]
        
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(r_c,np.real(func_c))
        ax.plot(r_c,np.imag(func_c))
        ax.vlines(r_c[f_max_ind],-f_max*1.1,f_max*1.1,colors='r')
        for ii in sub_tol_pts:
            ax.vlines(ii,-f_max*1.1,f_max*1.1,colors='grey')
        ax.vlines(r_c[sub_tol_inds[cutoff]],-f_max*1.1,f_max*1.1,colors='green')

        return None
    else:
        return r_c[sub_tol_inds[cutoff]]
    
    
    
def gen_const(val):
    
    '''
    
    Create executable function returning a constant value
    
    *args*:  

        - **val**: constant value to return when executable function
    
    *return*:   
       
        - lambda function with constant value
    
    ***
    '''
    
    return lambda x: val
    


# -*- coding: utf-8 -*-
#Created on Sat Oct 20 08:27:49 2018

#Created on Tue Mar 26 20:42:09 2019

#@author: ryanday


#MIT License

#Copyright (c) 2018 Ryan Patrick Day

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.



#FERMI SURFACE TRIANGULATION,
#USING THE TETRAHEDRAL MESH APPROACH, motivated by the discussion in Appendix A of
#Helmholtz Fermi Surface Harmonics: an efficient approach for treating anisotropic problems involving Fermi surface integrals
#May 2014New Journal of Physics
#DOI: 10.1088/1367-2630/16/6/063014




import chinook.tetrahedra as tetrahedra
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from operator import itemgetter
import warnings
warnings.filterwarnings("error")


    


def EF_tetra(TB,NK,EF,degen=False,origin=None):
    
    '''
    Generate a tetrahedra mesh of k-points which span the BZ with even distribution
    Diagonalize over this mesh and then compute the resulting density of states as
    prescribed in the above paper. 
    
    *args*:

        - **TB**: tight-binding model object
        
        - **NK**: int or list,tuple of 3 int indicating number of k-points in mesh
        
        - **EF**: float, Fermi energy, or energy of interest
        
    *kwargs*:

        - **degen**: bool, flag for whether the bands are two-fold degenerate, as for 
        Kramers degeneracy
        
        - **origin**: numpy array of 3 float,  corresponding to the desired centre of the 
        plotted Brillouin zone
        
    *return*:

        - **surfaces**: dictionary of dictionaries: Each key-value pair corresponds
        to a different band index. For each case, the value is a dictionary with key-value
        pairs:
            
            - *'pts'*: numpy array of Nx3 float, the N coordinates of EF crossing
            
            - *'tris'*: numpy array of Nx3 int, the triangulation of the surface
    
    ***
    '''
    
    kpts,tetra = tetrahedra.mesh_tetra(TB.avec,NK)
    if origin is not None:
        com = np.mean(kpts,axis=0)
        shift = origin-com
        kpts+=shift

    TB.Kobj.kpts = kpts
    TB.solve_H()
    if degen:
        inc = 2
        TB.Eband = TB.Eband[:,::inc]
        
    else:
        inc = 1
    surfaces = {bi:{'pts':[],'tris':[]} for bi in range(len(TB.basis[::inc])) if (TB.Eband[:,bi].min()<=EF and TB.Eband[:,bi].max()>=EF)} #iinitialize Fermi surface only for those bands which actually have one...

    for bi in surfaces.keys(): #iterate only over bands which have Fermi surface
        for ki in range(len(tetra)):
            E_tmp = TB.Eband[tetra[ki],bi] #vertex energies

            register = np.zeros((4,2))
            register[:,1] = tetra[ki]
            register[:,0] = E_tmp
            register = np.array(sorted(register,key=itemgetter(0))) #sort corners based on energies
            kinds = register[:,1].astype(int) #indices
            t = register[:,0] #energies

            if t[0]<=EF and EF<=t[3]: #if True, Fermi level cuts through this tetrahedra
                k = kpts[kinds] #get k-coordinates
                if t[0]<=EF<t[1]:
                    
                    t5 = k[0]+(EF-t[0])/(t[2]-t[0])*(k[2]-k[0])
                    t6 = k[0] + (EF-t[0])/(t[3]-t[0])*(k[3]-k[0])
                    t7 = k[0] + (EF-t[0])/(t[1]-t[0])*(k[1]-k[0])
                
                    tri = [np.around([t5,t6,t7],4)]
                    
                elif t[1]<=EF<t[2]:
                
                    t5 = k[0]+(EF-t[0])/(t[2]-t[0])*(k[2]-k[0])
                    t6 = k[0] + (EF-t[0])/(t[3]-t[0])*(k[3]-k[0])
                    t7 = k[1] +(EF-t[1])/(t[3]-t[1])*(k[3]-k[1])
                    t8 = k[1] + (EF-t[1])/(t[2]-t[1])*(k[2]-k[1])
                    tri = np.around(sim_tri([t5,t6,t7,t8]),4)
                
                elif t[2]<=EF<=t[3]:
                    
                    t5 = k[0] + (EF-t[0])/(t[3]-t[0])*(k[3]-k[0])
                    t6 = k[1] + (EF-t[1])/(t[3]-t[1])*(k[3]-k[1])
                    t7 = k[2] + (EF-t[2])/(t[3]-t[2])*(k[3]-k[2])
                    tri = [np.around([t5,t6,t7],4)]
                else:
                    continue
                for trii in tri:
                    coords = np.zeros(3,dtype=int)
                    for t in list(enumerate(trii)):
                        if len(surfaces[bi]['pts'])>0:
                            dv_bi = np.linalg.norm(t[1]-surfaces[bi]['pts'],axis=1) 
                        else: 
                            dv_bi=np.array([1])
                        if dv_bi.min()>1e-10:
                            surfaces[bi]['pts'].append(t[1])
                            coords[t[0]] = len(surfaces[bi]['pts'])-1
                        
                        
                        else:
                            coords[t[0]] =np.where(dv_bi==0)[0][0] 
                    surfaces[bi]['tris'].append(coords)
        surfaces[bi]['pts'] = np.array(surfaces[bi]['pts'])
        surfaces[bi]['tris'] = np.array(surfaces[bi]['tris'])

    
    return surfaces



def FS_generate(TB,Nk,EF,degen = False,origin=None,ax=None):
    
    '''

    Wrapper function for computing Fermi surface triangulation, and then plotting
    the result. 
    
    *args*:

        - **TB**: tight-binding model object
        
        - **Nk**: int, or tuple/list of 3 int, number of k-points in Brillouin zone mesh
        
        - **EF**: float, Fermi energy, or constant energy level of interest
        
    *kwargs*:

        - **degen**: bool, flag for whether the bands are two-fold degenerate, as for 
        Kramers degeneracy
        
        - **origin**: numpy array of 3 float,  corresponding to the desired centre of the 
        plotted Brillouin zone
        
        - **ax**: matplotlib Axes, option for plotting onto existing Axes

    
    *return*:

        - **surfaces**: dictionary of dictionaries: Each key-value pair corresponds
        to a different band index. For each case, the value is a dictionary with key-value
        pairs:
            
            - *'pts'*: numpy array of Nx3 float, the N coordinates of EF crossing
            
            - *'tris'*: numpy array of Nx3 int, the triangulation of the surface
            
        - **ax**: matplotlib Axes, for further modification

    ***
    '''
    
    surfaces = EF_tetra(TB,Nk,EF,degen,origin)
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111,projection='3d')
    for bi in surfaces:
        Fk=surfaces[bi]['pts']

        ax.plot_trisurf(Fk[:,0],Fk[:,1],Fk[:,2],cmap=cm.magma,triangles=surfaces[bi]['tris'],linewidth=0.1,edgecolor='w',alpha=1.0)
   # ax.grid(False)
#    ax.axis('off')
   # ax.set_aspect(1)
    return surfaces,ax


def heron(vert):
    '''
    Heron's algorithm for calculation of triangle area, defined by only the vertices
    
    *args*:

        - **vert**: numpy array of 3x3 indicating vertices of triangle
    
    *return*:

        - float, area of triangle
        
    ***
    '''
    
    L = [np.linalg.norm(vert[np.mod(j,3)]-vert[np.mod(j-1,3)]) for j in range(3)]
    S = 0.5*sum(L)
    
    try:
        area = np.sqrt(abs(S*(S-L[0])*(S-L[1])*(S-L[2])))
    except RuntimeWarning:
        print(S,L)
        area = 0
    return area


def sim_tri(vert):
    
    '''
    
    Take 4 vertices of a quadrilateral and split into two alternative triangulations of the corners.
    Return the vertices of the triangulation which has the more similar areas between the two
    triangles decomposed.
    
    *args*:

        - **vert**: (4 by 3 numpy array (or list) of float) in some coordinate frame 
        
    *return*:
    
        - **tris[0]** , **tris[1]**: two numpy arrays of size 3 by 3 float containing
        the coordinates of a triangulation 
    
    ***
    '''
    
    tris = np.array([[vert[0],vert[2],vert[1]],[vert[0],vert[3],vert[2]],[vert[0],vert[3],vert[1]], [vert[1],vert[3],vert[2]]])
    areas = [heron(ti) for ti in tris]
    if abs(areas[1]-areas[0])<abs(areas[3]-areas[2]):
        return [tris[0],tris[1]]
    else:
        return [tris[2],tris[3]]
    
    
#def FS_contour(TB,kmax,Ef):
#    kr = np.linspace(kmax/1000,kmax,100)
#    phi  = np.linspace(0,2*np.pi,100)
#    k_phi = np.zeros((len(phi),3))
#    for 
#    kvecs = np.array([[kr[i]*np.cos(phi[j]),kr[i]*np.sin(phi[j]),0] for j in range(len(phi)) for i in range(len(kr))])


#    TB.Kobj.kpts = kvecs
        
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#Created on Thu Nov  9 21:38:24 2017

#@author: ryanday


#MIT License

#Copyright (c) 2018 Ryan Patrick Day

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.
import numpy as np
import chinook.SlaterKoster as SK
import chinook.rotation_lib as rot_lib
import chinook.Ylm as Ylm


hb = 6.626*10**-34/(2*np.pi)
c  = 3.0*10**8
q = 1.602*10**-19
A = 10.0**-10
me = 9.11*10**-31
mn = 1.67*10**-27
kb = 1.38*10**-23



    

def txt_build(filename,cutoff,renorm,offset,tol):
    
    '''

    Build Hamiltonian from textfile, input is of form
    o1,o2,x12,xy12,z12,t12, output in form [o1,o2,x12,y12,z12,t12]
    
    *args*:

        - **filename**: string, name of file
        
        - **cutoff**: float, maximum distance of hopping allowed, Angstrom
        
        - **renorm**: float, renormalization of the bandstructure
        
        - **offset**: float, energy offset of chemical potential, electron volts
        
        - **tol**: float, minimum Hamiltonian matrix element amplitude
        
    *return*:

        - **Hlist**: the list of Hamiltonian matrix elements
    
    ***
    '''
    
    Hlist = []
    
    with open(filename,'r') as origin:
        
        for line in origin:
            
            spl = line.split(',')
            R = np.array([float(spl[2]),float(spl[3]),float(spl[4])])
            Hval = complex(spl[5])
            
            if len(spl)>6:
                Hval+=1.0j*float(spl[6])
            if abs(Hval)>tol and  np.linalg.norm(R)<cutoff:
                Hval*=renorm
                if np.linalg.norm(R)==0.0:
                    Hval-=offset
                    
                tmp = [int(spl[0]),int(spl[1]),R[0],R[1],R[2],Hval]

                Hlist.append(tmp)
            
    origin.close()
            
    return Hlist

def sk_build(avec,basis,Vdict,cutoff,tol,renorm,offset):
    
    '''

    Build SK model from using D-matrices, rather than a list of SK terms from table.
    This can handle orbitals of arbitrary orbital angular momentum in principal, 
    but right now implemented for up to and including f-electrons. 
    NOTE: f-hoppings require thorough testing
    
    *args*:

        - **avec**: numpy array 3x3 float, lattice vectors
        
        - **basis**: list of orbital objects
        
        - **Vdict**: dictionary, or list of dictionaries, of Slater-Koster integrals/ on-site energies
        
        - **cutoff**: float or list of float, indicating range where Vdict is applicable
        
        - **tol**: float, threshold value below which hoppings are neglected 
        
        - **offset**: float, offset for Fermi level 
        
    *return*:

        - **H_raw**: list of Hamiltonian matrix elements, in form [o1,o2,x12,y12,z12,t12]
    
    ***
    '''
        
    Vdict,cutoff,pts = cluster_init(Vdict,cutoff,avec) #build region of lattice points, containing at least the cutoff distance
    V = Vdict[0]
    if basis[0].spin!=basis[-1].spin: #only calculate explicitly for a single spin species
        brange = int(len(basis)/2)
    else:
        brange = len(basis)

    SK_matrices = SK.SK_full(basis[:brange]) #generate the generalized Slater-Koster matrices, as functions of R and potential V
    index_orbitals = index_ordering(basis[:brange]) #define the indices associated with the various orbital shells in the basis,
    H_raw = on_site(basis[:brange],V,offset) #fill in the on-site energies

    for i1 in index_orbitals:
        for i2 in index_orbitals:
            if index_orbitals[i1][index_orbitals[i1]>-1].min()<=index_orbitals[i2][index_orbitals[i2]>-1].min():
                o1o2 = (i1[0],i2[0],i1[1],i2[1],i1[2],i2[2])
                R12 = (np.array(i2[3:6])-np.array(i1[3:6])) 
                SKmat = SK_matrices[o1o2]
        
                for p in pts: #iterate over the points in the cluster
                    Rij = R12 + np.dot(p,avec)
                    Rijn = np.linalg.norm(Rij) #compute norm of the vector
#                    
                    if 0<Rijn<cutoff[-1]: #only proceed if within the cutoff distance
                        V = Vdict[np.where(Rijn>=cutoff)[0][-1]]
                    
                        Vlist = Vlist_gen(V,o1o2)
                        if Vlist is None:
                            continue
                        elif len(Vlist)==0:
                            continue
                        Euler_A,Euler_B,Euler_y = rot_lib.Euler(rot_lib.rotate_v1v2(Rij,np.array([0,0,1])))
                    
                        SKvals = mirror_SK([vi for vi in Vlist])
                        SKmat_num = SKmat(Euler_A,Euler_B,Euler_y,SKvals) #explicitly compute the relevant Hopping matrix for this vector and these shells
                        if abs(SKmat_num).max()>tol:

                            append = mat_els(Rij,SKmat_num,tol,index_orbitals[i1],index_orbitals[i2])
                            H_raw = H_raw + append
    return H_raw #finally return the list of Hamiltonian matrix elements




def on_site(basis,V,offset):
    
    '''

    On-site matrix element calculation. Try both anl and alabel formats,
    if neither is defined, default the onsite energy to 0.0 eV
    
    *args*:

        - **basis**: list of orbitals defining the tight-binding basis
        
        - **V**: dictionary, Slater Koster terms
        
        - **offset**: float, EF shift
        
    *return*:

        - **Ho**: list of Hamiltonian matrix elements
        
    ***
    '''
    Ho = []
    for oi in basis:
        try:
            H = V['{:d}{:d}{:d}'.format(oi.atom,oi.n,oi.l)]
        except KeyError:
            try:
                H = V['{:d}{:s}'.format(oi.atom,oi.label)]
            except KeyError:
                H = 0.0
        Ho.append([oi.index,oi.index,0.0,0.0,0.0,float(H-offset)])
    return Ho
    
                
def mat_els(Rij,SKmat,tol,i1,i2):
    '''

    Extract the pertinent, and non-zero elements of the Slater-Koster matrix
    and transform to the conventional form of Hamiltonian list entries
    (o1,o2,Rij0,Rij1,Rij2,H12(Rij))
    
    *args*:

        - **Rij**: numpy array of 3 float, relevant connecting vector 
        
        - **SKmat**: numpy array of float, matrix of hopping elements 
        for the coupling of two orbital shells 
        
        - **tol**: float, minimum hopping included in model
        
        - **i1**, **i2**: int,int, proper index ordering for the relevant
        instance of the orbital shells involved in hopping
        
    *return*:
        
        - **out**: list of Hamiltonian matrix elements, extracted from the
        ordered SKmat, in form [[o1,o2,x12,y12,z12,H12],...]
    
    ***    
    '''
    inds = np.where(abs(SKmat)>tol)
    out = []
    for ii in range(len(inds[0])):
        i_1 = i1[inds[0][ii]]
        i_2 = i2[inds[1][ii]]
        
        if -1<i_1<=i_2:
            out.append([i_1,i_2,*Rij,SKmat[inds[0][ii],inds[1][ii]]])

    return out
                    

def index_ordering(basis):
    
    '''

    We use an universal ordering convention for defining the Slater-Koster matrices
    which may (and most likely will) not match the ordering chosen by the user.
    To account for this, we define a dictionary which gives the ordering, relative 
    to the normal order convention defined here, associated with a given a-n-l shell
    at each site in the lattice basis.
    
    *args*:

        - **basis**: list of orbital objects
    
    *return*:

        - **indexing**: dictionary of key-value pairs (a,n,l,x,y,z):numpy.array([...])
    
    ***
    '''
    normal_order = {0:{'':0},1:{'x':0,'y':1,'z':2},2:{'xz':0,'yz':1,'xy':2,'ZR':3,'XY':4},3:{'z3':0,'xz2':1,'yz2':2,'xzy':3,'zXY':4,'xXY':5,'yXY':6}}
    indexing = {}
    for b in basis:
        anl = (b.atom,b.n,b.l,*np.around(b.pos,4))
        if anl not in indexing.keys():
            indexing[anl] = -1*np.ones(2*b.l+1)
        indexing[anl][normal_order[b.l][b.label[2:]]] = b.index
        
    return indexing

      

def Vlist_gen(V,pair):
   
    '''

    Select the relevant hopping matrix elements to be used in defining the value
    of the Slater-Koster matrix elements for a given pair of orbitals. Handles situation where
    insufficient parameters have been passed to system.
    
    *args*:

        - **V**: dictionary of Slater-Koster hopping terms
        
        - **pair**: tuple of int defining the orbitals to be paired, (a1,a2,n1,n2,l1,l2)
    
    *return*:
        
        - **Vvals**: numpy array of Vllx related to a given pairing, e.g. for s-p np.array([Vsps,Vspp])
    
    ***
    '''
    order = {'S':0,'P':1,'D':2,'F':3,0:'S',1:'P',2:'D',3:'F'}
    vstring = '{:d}{:d}{:d}{:d}{:d}{:d}'.format(*pair[:6])
    l = max(pair[4],pair[5])
    if len(V.keys())<(l+1):
        print('WARNING, insufficient number of Slater-Koster parameters passed: filling missing values with zeros.')
        for l_index in range(l+1):
            hopping_type = vstring+order[l_index]
            if hopping_type not in V.keys():
                V[hopping_type] = 0
    try:
        Vkeys = np.array(sorted([[l-order[vi[-1]],vi] for vi in V if vi[:-1]==vstring]))[:,1]
        Vvals = np.array([V[vk] for vk in Vkeys])
    except IndexError:
        vstring = '{:d}{:d}{:d}{:d}{:d}{:d}'.format(pair[1],pair[0],pair[3],pair[2],pair[5],pair[4])
        try: 
            Vkeys = np.array(sorted([[l-order[vi[-1]],vi] for vi in V if vi[:-1]==vstring]))[:,1]
            pre = (-1)**(pair[4]+pair[5]) #relative parity of the two coupled states
            Vvals = pre*np.array([V[vk] for vk in Vkeys])

        except IndexError:
            return None
    return Vvals
    
    
            
def mirror_SK(SK_in):
    
    '''
    
    Generate a list of values which is the input appended with its mirror 
    reflection. The mirror boundary condition suppresses the duplicate of the
    last value. e.g. [0,1,2,3,4] --> [0,1,2,3,4,3,2,1,0], 
    ['r','a','c','e','c','a','r'] --> ['r','a','c','e','c','a','r','a','c','e','c','a','r']
    Intended here to take an array of Slater-Koster hopping terms and reflect about 
    its last entry i.e. [Vsps,Vspp] -> [Vsps,Vspp,Vsps]
    
    *args*:

        - **SK_in**: iterable, of arbitrary length and data-type
        
    *return*:

        - list of values with same data-type as input
    
    ***
    '''
    return list(SK_in) + (SK_in[-2::-1])



def cluster_init(Vdict,cutoff,avec):
    
    '''
    Generate a cluster of neighbouring lattice points to use
    in defining the hopping paths--ensuring that it extends
    sufficiently far enough to capture even the largest hopping vectors.
    Also reforms the SK dictionary and cutoff lengths to be in list format.
    Returns an array of lattice points which go safely to the edge of the cutoff range.
    
    *args*:

        - **Vdict**: dictionary, or list of dictionaries of Slater Koster matrix elements
        
        - **cutoff**: float, or list of float
        
        - **avec**: numpy array of 3x3 float
        
    *return*:

        - **Vdict**: list of length 1 if a single dictionary passed, else unmodified
        
        - **cutoff**: numpy array, append 0 to the beginning of the cutoff list,
        else leave it alone.
        
        - **pts**: numpy array of lattice vector indices for a region of lattice points around
        the origin.
        
    ***
    '''
    if isinstance(cutoff,(int,float)) and not isinstance(cutoff,bool):
        cutoff = np.array([0.0,cutoff])
        Vdict = [Vdict]
    else:
        
    
        if cutoff[0]>0:
            cutoff.insert(0,0)
            cutoff = np.array(cutoff)
        else:
            cutoff = np.array(cutoff)
        

    pt_max = np.ceil(np.array([(cutoff).max()/np.linalg.norm(avec[i]) for i in range(len(avec))]).max())
    pts = region(int(pt_max)+1)
    return Vdict,cutoff,pts



###############################################################################
#########################Spin Orbit Coupling###################################
###############################################################################
    

def spin_double(H,lb):
    '''
    Duplicate the kinetic Hamiltonian terms to extend over the spin-duplicated 
    orbitals, which are by construction in same order and appended to end of the
    original basis.
    
    *args*:

        - **H**: list, Hamiltonian matrix elements [[o1,o2,x,y,z,H12],...]
        
        - **lb**: int, length of basis before spin duplication
        
    *return*:

        - **h2** modified copy of **H**, filled with kinetic terms for both 
        spin species
    
    ***
    '''
    lenb = int(lb/2)
    h2 = []
    for i in range(len(H)):
        h2.append([H[i][0]+lenb,H[i][1]+lenb,H[i][2],H[i][3],H[i][4],H[i][5]])
    return h2


def SO(basis):

    '''
    Generate L.S  matrix-elements for a given basis. 
    This is generic to all l, except the normal_order, which is defined here up to 
    and including the f electrons.
    Otherwise, this method is generic to any orbital angular momentum.
    
    In the factors dictionary defined here indicates the weight of the 
    different :math:`L_iS_i` terms. The keys are tuples of (L+/-/z,S+/-/z)
    in a bit of a cryptic way: for L, (0,1,2) ->(-1,0,1) and
    for S, (-1,0,1) = S1-S2 with S1,2 = +/- 1 here
    
    L+,L-,Lz matrices are defined for each l shell in the basis, 
    transformed into the basis of the tight-binding model.
    The nonzero terms will then just be used along with the spin and
    weighted by the factor value, and slotted into a len(**basis**)xlen(**basis**) matrix **HSO**
    
    *args*:

        - **basis**: list of orbital objects
    
    *return*:

        - **HSO**: list of matrix elements in standard format [o1,o2,0,0,0,H12]
        
    ***
    '''
    Md = Ylm.Yproj(basis)
    normal_order = {0:{'':0},1:{'x':0,'y':1,'z':2},2:{'xz':0,'yz':1,'xy':2,'ZR':3,'XY':4},3:{'z3':0,'xz2':1,'yz2':2,'xzy':3,'zXY':4,'xXY':5,'yXY':6}}
    factors = {(2,-1):0.5,(0,1):0.5,(1,0):1.0}
    L,al = {},[]
    HSO = []
    for o in basis[:int(len(basis)/2)]:
        if (o.atom,o.n,o.l) not in al:
            al.append((o.atom,o.n,o.l))
            Mdn = Md[(o.atom,o.n,o.l,-1)]
            Mup = Md[(o.atom,o.n,o.l,1)]
            Mdnp = np.linalg.inv(Mdn)
            Mupp = np.linalg.inv(Mup)
            L[(o.atom,o.n,o.l)] = [np.dot(Mupp,np.dot(Lm(o.l),Mdn)),np.dot(Mupp,np.dot(Lz(o.l),Mup)),np.dot(Mdnp,np.dot(Lp(o.l),Mup))]

    for o1 in basis:
        for o2 in basis:
            if o1.index<=o2.index:
                LS_val = 0.0
                if np.linalg.norm(o1.pos-o2.pos)<0.0001 and o1.l==o2.l and o1.n==o2.n:
                    inds = (normal_order[o1.l][o1.label[2:]],normal_order[o2.l][o2.label[2:]])
                    
                    ds = (o1.spin-o2.spin)/2.
                    if ds==0:
                        s=0.5*np.sign(o1.spin)
                    else:
                        s=1.0
                    for f in factors:
                        if f[1]==ds:
                            LS_val+=o1.lam*factors[f]*L[(o1.atom,o1.n,o1.l)][f[0]][inds]*s
                    HSO.append([o1.index,o2.index,0.,0.,0.,LS_val])

    return HSO


def Lp(l):

    '''
    L+ operator in the :math:`l`, :math:`m_l` basis, organized with 
    (0,0) = |l,l>... (2l,2l) = |l,-l>
    The nonzero elements are on the upper diagonal
    
    *arg*:

        - **l**: int orbital angular momentum
    
    *return*:

        - **M**: numpy array (2l+1,2l+1) of real float
        
    ***
    '''
    M = np.zeros((2*l+1,2*l+1))
    r = np.arange(0,2*l,1)
    M[r,r+1]=1.0
    vals = [0]+[np.sqrt(l*(l+1)-(l-m)*(l-m+1)) for m in range(1,2*l+1)]
    M = M*vals
    return M

def Lm(l):

    '''
    L- operator in the l,m_l basis, organized with 
    (0,0) = |l,l>... (2l,2l) = |l,-l>

    The nonzero elements are on the upper diagonal
    
    *arg*:

        - **l**: int orbital angular momentum
    
    *return*:

        - **M**: numpy array (2l+1,2l+1) of real float
        
    ***
    '''
    M = np.zeros((2*l+1,2*l+1))
    r = np.arange(1,2*l+1,1)
    M[r,r-1]=1.0
    vals = [np.sqrt(l*(l+1)-(l-m)*(l-m-1)) for m in range(0,2*l)]+[0]
    M = M*vals
    return M

def Lz(l):

    '''
    Lz operator in the l,:math:`m_l` basis
    
    *arg*:

        - **l**: int orbital angular momentum
    
    *return*:

        - numpy array (2*l+1,2*l+1)

    ***
    '''
    return np.identity(2*l+1)*np.array([l-m for m in range(2*l+1)])



def AFM_order(basis,dS,p_up,p_dn):
  
  '''
  Add antiferromagnetism to the tight-binding model, by adding a different on-site energy to 
  orbitals of different spin character, on the designated sites. 
  
  *args*:

      - **basis**: list, orbital objects
      
      - **dS**: float, size of spin-splitting (eV)
      
      - **p_up**, **p_dn**: numpy array of float indicating the orbital positions
      for the AFM order
      
  *return*: 

      - **h_AF**: list of matrix elements, as conventionally arranged [[o1,o2,0,0,0,H12],...]

  ***
  '''
  h_AF = []
  for bi in basis:
      if np.linalg.norm(bi.pos-p_up)==0:
          if bi.spin<0:
              h_AF.append([bi.index,bi.index,0,0,0,dS])
          else:
              h_AF.append([bi.index,bi.index,0,0,0,-dS])
      elif np.linalg.norm(bi.pos-p_dn)==0:
          if bi.spin<0:
              h_AF.append([bi.index,bi.index,0,0,0,-dS])
          else:
              h_AF.append([bi.index,bi.index,0,0,0,dS])
  return h_AF
    
    
def FM_order(basis,dS):
    
    '''

     Add ferromagnetism to the system. Take dS to assume that the splitting puts 
     spin-up lower in energy by dS,and viceversa for spin-down. This directly
     modifies the *TB_model*'s **mat_els** attribute
     
     *args*:

         - **basis**: list, of orbital objects in basis
         
         - **dS**: float, energy of the spin splitting (eV)
    
     *return*:

         - list of matrix elements [[o1,o2,0,0,0,H12],...]
    
     ***
     '''
    return [[bi.index,bi.index,0,0,0,-np.sign(bi.spin)*dS] for bi in basis]


#def Efield(basis,field,orbital_type='Slater'):
    
    '''
    Define a set of matrix elements which introduce an electric field, treated at the level of a dipole operator.
    
    TODO

    '''
 #   return None
    
    



def region(num):
    
    '''

    Generate a symmetric grid of points in number of lattice vectors. 
    
    *args*:

        - **num**: int, grid will have size 2*num+1 in each direction
    
    *return*:

        - numpy array of size ((2*num+1)**3,3) with centre value of first entry
        of (-num,-num,-num),...,(0,0,0),...,(num,num,num)

    ***
    '''
    num_symm = 2*num+1
    return np.array([[int(i/num_symm**2)-num,int(i/num_symm)%num_symm-num,i%num_symm-num] for i in range((num_symm)**3)])





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#Created on Thu Nov  9 21:21:15 2017
#@author: ryanday
#MIT License

#Copyright (c) 2018 Ryan Patrick Day

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

import numpy as np
import chinook.atomic_mass as am
from chinook.wigner import WignerD
import chinook.rotation_lib as rotlib
from operator import itemgetter


##STANDARD ORBITAL PROJECTIONS IN THE BASIS OF SPHERICAL HARMONICS Y_LM
projdict={"0":np.array([[1.0,0.0,0.0,0.0]]),
               "1x":np.array([[-np.sqrt(0.5),0.0,1,1],[np.sqrt(0.5),0.0,1,-1]]),"1y":np.array([[0.0,np.sqrt(0.5),1,1],[0,np.sqrt(0.5),1,-1]]),"1z":np.array([[1,0,1,0]]),
                "2xy":np.array([[0.0,-np.sqrt(0.5),2,2],[0.0,np.sqrt(0.5),2,-2]]),"2yz":np.array([[0.0,np.sqrt(0.5),2,1],[0.0,np.sqrt(0.5),2,-1]]),
                "2xz":np.array([[-np.sqrt(0.5),0,2,1],[np.sqrt(0.5),0,2,-1]]),"2ZR":np.array([[1.0,0.0,2,0]]),"2XY":np.array([[np.sqrt(0.5),0,2,2],[np.sqrt(0.5),0,2,-2]]),
                "3z3":np.array([[1.0,0.0,3,0]]),"3xz2":np.array([[np.sqrt(0.5),0,3,-1],[-np.sqrt(0.5),0,3,1]]),
                "3yz2":np.array([[0,np.sqrt(0.5),3,-1],[0,np.sqrt(0.5),3,1]]),"3xzy":np.array([[0,-np.sqrt(0.5),3,2],[0,np.sqrt(0.5),3,-2]]),
                "3zXY":np.array([[np.sqrt(0.5),0,3,2],[np.sqrt(0.5),0,3,-2]]),"3xXY":np.array([[-np.sqrt(0.5),0,3,3],[np.sqrt(0.5),0,3,-3]]),
                "3yXY":np.array([[0,np.sqrt(0.5),3,3],[0,np.sqrt(0.5),3,-3]])}
                

##PHYSICAL CONSTANTS##
hb = 6.626*10**-34/(2*np.pi)
c  = 3.0*10**8
q = 1.602*10**-19
A = 10.0**-10
me = 9.11*10**-31
mN = 1.67*10**-27
kb = 1.38*10**-23

class orbital:
    '''
    The **orbital** object carries all essential details of the elements of the
    model Hamiltonian basis, for both generation of the tight-binding model, in
    addition to the evaluation of expectation values and ARPES intensity.
    
    '''
    
    def __init__(self,atom,index,label,pos,Z,orient=[0.0],spin=1,lam=0.0,sigma=1.0,slab_index=None):
        
        '''
        Initialize the **orbital** object.
        
        *args*:

            - **atom**: int, index of inequivalent atoms in the basis
            
            - **index**: int, index in the list of orbitals which consitute the 
            basis
            
            - **label**: string, should be of format nlXX w/ lXX the orbital convention at top of *orbital.py*
        
            - **pos**: numpy array of 3 float, position of orbital in direct space, units of Angstrom
            
            - **Z**: int, atomic number
            
        *kwargs*:

            - **orient**: various possible format. Can pass a projection array, 
            as in case of **projdict**, each element being length 4 array of 
            [Re(proj),Im(proj), l,m]. Alternatively, can pass a conventional 
            orbital label, and define a rotation: this is passed as a list, with 
            first element a numpy array of 3 float indicating the rotation axis, 
            and the second element a float, corresponding to angle, in radian,
            of the rotation
                
            - **spin**: int, +/-1, indicating the spin projection, default to 
            1 for spinless calculations
                
            - **lam**: float, strength of spin-orbit coupling, eV
                
            - **sigma**: float, effective scattering cross section, a.u.
                
            - **slab_index**: int, index in ordered list of slab orbitals, unused for bulk
            calculations
            
        ***
        '''
        
        
        self.atom = atom
        self.index = index
        self.label = label
        self.pos = pos 
        self.Z = Z 
        self.mn = am.get_mass_from_number(self.Z)*mN #atomic mass
        self.spin = spin #spin (+/-1)
        self.lam = lam #spin-orbit coupling strength
        self.sigma = sigma #scattering cross section, can be reduced/amplified for a given orbital
        self.n,self.l = int(self.label[0]),int(self.label[1])
        self.Dmat = np.identity(2*self.l+1)
        if slab_index is None:
            self.slab_index = index #this is redundant for bulk calc, but index in slab is distinct from lattice index
            self.depth = 0.0
        else:
            self.slab_index = slab_index
            self.depth = self.pos[2]

        self.orient = orient
        if type(self.orient)==np.ndarray:
            self.proj = self.orient #projection into Ylm of form in projdict--array: [Real,Imag,l,ml]
        elif type(self.orient)==list: #can also pass a rotation from the conventional orientation
            self.proj = projdict[self.label[1:]]
            if abs(self.orient[-1])>0:
                self.proj,self.Dmat = rot_projection(self.l,self.proj,self.orient)

        else:
            self.proj = projdict[self.label[1:]]
            
    def copy(self):

        '''
        Copy by value method for orbital object
        
        *return*:

            - **orbital_copy**: duplicate of **orbital** object
            
        ***
        '''
        orbital_copy = orbital(self.atom,self.index,self.label,self.pos,self.Z,self.orient,self.spin,self.lam,self.sigma,self.slab_index)
        orbital_copy.proj = self.proj.copy()
        orbital_copy.Dmat = self.Dmat.copy()
        return orbital_copy

        
def rot_projection(l,proj,rotation):
    
    '''
    Go through a projection array, and apply the intended transformation to
    the Ylm projections in order.
    Define Euler angles in the z-y-z convention
    THIS WILL BE A COUNTERCLOCKWISE ROTATION ABOUT a vector BY angle gamma 
    expressed in radians. Note that we always define spin in the lab-frame, so
    spin degrees of freedom are not rotated when we rotate the orbital degrees
    of freedom.
    
    *args*:

        - **l**: int,orbital angular momentum
        
        - **proj**: numpy array of shape Nx4 of float, each element is
        [Re(projection),Im(projection),l,m]
        
        - **rotation**: float, or list, defining rotation of orbital. If float,
        assume rotation about z-axis. If list, first element is a numpy array 
        of len 3, indicating rotation vector, and second element is float, angle.
        
    *return*:

        - **proj**: numpy array of Mx4 float, as above, but modified, and may
        now include additional, or fewer elements than input *proj*.

        - **Dmat**: numpy array of (2l+1)x(2l+1) complex float indicating the 
        Wigner Big-D matrix associated with the rotation of this orbital shell
        about the intended axis.
    
    ***
    '''
        
    if len(rotation)==1:
        rotation = (np.array([0,0,1]),rotation)
    Ylm_vec = np.zeros((2*l+1),dtype=complex)
    for a in range(len(proj)):
        Ylm_vec[int(proj[a,-1]+l)] +=proj[a,0]+1.0j*proj[a,1]
    
    A,B,y = rotlib.Euler(rotation)
    Dmat = WignerD(l,A,B,y)
    Ynew = np.dot(Dmat,Ylm_vec)

    proj = []

    for a in range(2*l+1):
        if abs(Ynew[a])>10**-10:
            proj.append([np.around(np.real(Ynew[a]),10),np.around(np.imag(Ynew[a]),10),l,a-l])
            
    proj = np.array(proj)
    return proj,Dmat

def fact(n):

    '''
    Recursive factorial function, works for any non-negative integer.
    
    *args*:

        - **n**: int, or integer-float
        
    *return*:

        - int, recursively evaluates the factorial of the initial input value.
    
    ***
    '''
    if (np.floor(n)-n)>0:
        print('Error, non-integer value passed to factorial.')
        return 0
    if n<0:
        print('negative factorial!')
        return 0
    elif n==0:
        return 1
    else:
        return n*fact(n-1)
    
    
def slab_basis_copy(basis,new_posns,new_inds):
    
    '''
    Copy elements of a slab basis into a new list of
    orbitals, with modified positions and index ordering.
    
    *args*:

        - **basis**: list or orbital objects
        
        - **new_posns**: numpy array of len(basis)x3 float, new positions for
        orbital
        
        - **new_inds**: numpy array of len(basis) int, new indices for orbitals
        
    *return*:

        - **new_basis**: list of duplicated orbitals following modification.
    
    ***
    '''
    new_basis = np.empty(len(basis),dtype=orbital)
    for o in list(enumerate(basis)):
        local_orb_copy = o[1].copy()
    
        
        local_orb_copy.slab_index = int(new_inds[o[0]])
        local_orb_copy.index = o[0]
        local_orb_copy.pos = new_posns[o[0]]
        new_basis[int(new_inds[o[0]])] = local_orb_copy
    return new_basis
    

def sort_basis(basis,slab):
    
    '''
    Utility script for organizing an orbital basis that is out of sequence
    
    *args*:

        - **basis**: list of orbital objects
        
        - **slab**: bool, True or False if this is for sorting a slab
        
    *return*:

        - **orb_basis**: list of sorted orbital objects (by orbital.index value)
        
    ***
    '''
    inds = [[b[0],b[1].index] for b in list(enumerate(basis))]
    
    ind_sort = sorted(inds,key=itemgetter(1))

    orb_basis = [basis[i[0]].copy() for i in ind_sort]
    
    return orb_basis
    
def spin_double(basis,lamdict):

    '''
    Double the size of a basis to introduce spin to the problem.
    Go through the basis and create an identical copy with opposite spin and 
    incremented index such that the orbital basis order puts spin down in the first
    N/2 orbitals, and spin up in the second N/2.
    
    *args*:

        - **basis**: list of orbital objects
       
        - **lamdict**: dictionary of int:float pairs providing the
        spin-orbit coupling strength for the different inequivalent atoms in 
        basis.
    
    *return*:
    
        - doubled basis carrying all required spin information
        
    ***
    '''
    LB = len(basis)
    basis_double = []

    for ind in range(LB):
        basis[ind].spin = -1
        basis[ind].lam = lamdict[basis[ind].atom]
        spin_up = basis[ind].copy()
        spin_up.spin = 1
        spin_up.index = basis[ind].index+LB
        basis_double.append(spin_up)
    return basis + basis_double

            
if __name__=="__main__":
    
    Z = 26
    i = 0
    label = ["32xz","32yz","32xy"]
    pos = np.zeros(3)
    
    dxz = orbital(0,i,label[0],pos,Z)
    dyz = orbital(0,i+1,label[1],pos,Z)
    dxy = orbital(0,i+2,label[2],pos,Z)
    
    v = np.array([1,0,0]) 
    a = np.pi/2
    print('x')
    print(np.around(dxz.proj,3))
    proj_x,_ = rot_projection(2,dxz.proj,[v,a])
    print(np.around(proj_x,3))
  
        


# -*- coding: utf-8 -*-
#Created on Fri Aug 31 16:02:49 2018

#@author: ryanday
#MIT License

#Copyright (c) 2018 Ryan Patrick Day

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.



import sys
import numpy as np
from scipy.special import erf
import matplotlib.pyplot as plt
import chinook.tetrahedra as tetrahedra



def dos_broad(TB,NK,NE=None,dE=None,origin = np.zeros(3)):
    '''
    Energy-broadened discrete density of states calculation.
    The Hamiltonian is diagonalized over the kmesh defined by NK and
    states are summed, as energy-broadened Gaussian peaks, rather than
    delta functions. 
    
    *args*:

        - **TB**: tight-binding model object
        
        - **NK**: int, or tuple of int, indicating number of k-points
        
    *kwargs*:

        - **NE**: int, number of energy bins for final output
        
        - **dE**: float, energy broadening of peaks, eV
        
        - **origin**: numpy array of 3 float, indicating the origin of the mesh to be used,
        relevant for example in kz-specific contributions to density of states
        
    *return*:

        - **DOS**: numpy array of float, density-of-states in states/eV
        
        - **Elin**: numpy array of float, energy domain in eV
    
    ***
    '''
    kpts = tetrahedra.gen_mesh(TB.avec,NK)+origin
    TB.Kobj.kpts = kpts
    TB.solve_H()
    print('Diagonalization complete')
    if dE is None:
        dE = def_dE(TB.Eband)

    Elin = np.arange(TB.Eband.min()-10*dE,TB.Eband.max()+10*dE,dE*0.5)
    DOS = np.zeros(len(Elin))
    for ki in range(len(kpts)):
        sys.stdout.write('\r'+progress_bar(ki+1,len(kpts)))
        for bi in range(len(TB.basis)): #iterate over all bands

            tmp_add =  gaussian(TB.Eband[ki,bi],Elin,dE)
            DOS+=tmp_add
    print('\n')
    DOS = DOS/len(kpts)
    if NE is not None:
        E_resample = np.linspace(Elin[0],Elin[-1],NE)
        DOS_resample = np.interp(E_resample,Elin,DOS)
        DOS = DOS_resample
        Elin = E_resample
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.plot(Elin,DOS)
    
    return DOS,Elin


def def_dE(Eband):
    
    '''

    If energy broadening is not passed for density-of-states calculation,
    compute a reasonable value based on the energy between adjacent energies
    in the tight-binding calculation
    
    *args*:

        - **Eband**: numpy array of float, indicating band energies
        
    *return*:

        **dE**: float, energy broadening, as the smallest average energy spacing
        over all bands.
    
    ***
    '''
    
    dE = Eband.max()-Eband.min()
    for bi in range(np.shape(Eband)[1]):
        diff = abs(Eband[1:,bi]-Eband[:-1,bi]).mean()
        if dE>diff:
            dE = diff
    print('Broadening: {:0.04f} eV\n'.format(dE))
    return dE

def ne_broad_numerical(TB,NK,NE=None,dE=None,origin=np.zeros(3)):
    '''
    Occupation function, as a numerical integral over the density of states function.
    
    
    *args*:

        - **TB**: tight-binding model object
        
        - **NK**: int, or tuple of int, indicating number of k-points
        
    *kwargs*:

        - **NE**: int, number of energy bins for final output
        
        - **dE**: float, energy spacing of bins, in eV
        
        - **origin**: numpy array of 3 float, indicating the origin of the mesh to be used,
        relevant for example in kz-specific contributions to density of states
        
    *return*:

        - **ne**: numpy array of float, integrated density-of-states at each energy
        
        - **Elin**: numpy array of float, energy domain in eV
    
    ***
    '''
    DOS,Elin = dos_broad(TB,NK,NE,dE,origin)
    delta = Elin[1]-Elin[0]
    ne = np.zeros(len(Elin))
    for ii in range(len(Elin)):
        if ii==0:
            val = (2*DOS[ii]+DOS[ii+1])/3
        elif ii==(len(Elin)-1):
            val = (2*DOS[ii] + DOS[ii-1])/3
            ne[ii] = ne[ii-1]
        else:
            val = (DOS[ii-1]+DOS[ii]*2 + DOS[ii+1])/4
            ne[ii] = ne[ii-1]
        ne[ii] += val*delta
    return ne,Elin


def find_EF_broad_dos(TB,NK,occ,NE=None,dE=None,origin=np.zeros(3)):
    '''
    Find the Fermi level of a model Hamiltonian, for a designated electronic
    occupation. Note this is evaluated at T=0, so EF is well-defined.

    *args*:

        - **TB**: tight-binding model object
        
        - **NK**: int, or tuple of int, indicating number of k-points
        
        - **occ**: float, desired electronic occupation
        
    *kwargs*:

        - **NE**: int, number of energy bins for final output
        
        - **dE**: float, energy spacing of bins, in eV
        
        - **origin**: numpy array of 3 float, indicating the origin of the mesh to be used,
        relevant for example in kz-specific contributions to density of states
        
    *return*:

        - **EF**: float, Fermi level in eV

    ***
    '''
    
    ne,Elin = ne_broad_numerical(TB,NK,NE,dE,origin)
    ind_EF = np.where(ne>occ)[0][0]
    EF = Elin[ind_EF]
    print('Fermi Energy is at: {:0.05f}+/-{:0.06f} eV'.format(EF,Elin[1]-Elin[0]))
    return EF
    
        
        
def ne_broad_analytical(TB,NK,NE=None,dE=None,origin=np.zeros(3),plot = True):
    '''

    Analytical evaluation of the occupation function. Uses scipy's errorfunction
    executable to evaluate the analytical form of a Gaussian-broadened state's contribution
    to the total occupation, at each energy
    
    *args*:

        - **TB**: tight-binding model object
        
        - **NK**: int, or tuple of int, indicating number of k-points
        
    *kwargs*:

        - **NE**: int, number of energy bins for final output
        
        - **dE**: float, energy spacing of bins, in eV
        
        - **origin**: numpy array of 3 float, indicating the origin of the mesh to be used,
        relevant for example in kz-specific contributions to density of states
        
        - **plot**: boolean, default to True, if false, suppress plot output
        
    *return*:

        - **nE**: numpy array of float, occupied states
        
        - **Elin**: numpy array of float, energy domain in eV

    ***
    '''
    kpts = tetrahedra.gen_mesh(TB.avec,NK)+origin #just a mesh over the Brillouin zone
    TB.Kobj.kpts = kpts
    TB.solve_H()
    print('Diagonalization complete')

    if dE is None:
        dE = def_dE(TB.Eband)
    Elin = np.arange(TB.Eband.min()-10*dE,TB.Eband.max()+10*dE,dE*0.5)

    nE = np.zeros(len(Elin))
    for ki in range(len(kpts)):
        sys.stdout.write('\r'+progress_bar(ki+1,len(kpts)))
        for bi in range(len(TB.basis)): #iterate over all bands

            tmp_add =  error_function(TB.Eband[ki,bi],Elin,dE)
            nE+=tmp_add
    print('\n')
    nE /=len(kpts)
    if NE is not None:
        E_resample = np.linspace(Elin[0],Elin[-1],NE)
        nE_resample = np.interp(E_resample,Elin,nE)
        Elin = E_resample
        nE = nE_resample
    if plot:
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(Elin,nE)

    
    return nE,Elin


    
def error_function(x0,x,sigma):
    
    '''
    Integral over the gaussian function, evaluated from -infinity to x, using
    the scipy implementation of the error function
    
    *args*:

        - **x0**: float, centre of Gaussian, in eV
        
        - **x**: numpy array of float, energy domain eV
        
        - **sigma**: float, width of Gaussian, in eV
        
    *return*:

        - analytical form of integral
    
    ***
    '''
    return 0.5*(erf((x-x0)/(np.sqrt(2)*sigma))+1)      
            
            
            
def gaussian(x0,x,sigma):

    '''
    Evaluate a normalized Gaussian function.
    
    *args*:

        - **x0**: float, centre of peak, in eV
        
        - **x**: numpy array of float, energy domain in eV
        
        - **sigma**: float, width of Gaussian, in eV

    *return*:

        - numpy array of float, gaussian evaluated.

    ***
    '''
    return np.exp(-(x-x0)**2/(2*sigma**2))*np.sqrt(1/(2*np.pi))/sigma
################# Density of States following the Blochl Prescription #######################
###############https://journals.aps.org/prb/pdf/10.1103/PhysRevB.49.16223####################
    
def dos_tetra(TB,NE,NK):
    '''

    Generate a tetrahedra mesh of k-points which span the BZ with even distribution
    Diagonalize over this mesh and then compute the resulting density of states as
    prescribed in the above paper. 
    The result is plotted, and DOS returned
    
    *args*:

        - **TB**: tight-binding model object
        
        - **NE**: int, number of energy points
        
        - **NK**: int or list of 3 int -- number of k-points in mesh
        
    *return*:

        - **Elin**: linear energy array of float, spanning the range of the eigenspectrum
        
        - **DOS**: numpy array of float, same length as Elin, density of states
    
    ***
    '''
    kpts,tetra = tetrahedra.mesh_tetra(TB.avec,NK)
    print('k tetrahedra defined')
    TB.Kobj.kpts = kpts
    TB.solve_H()
    print('Diagonalization complete')
    Elin = np.linspace(TB.Eband.min(),TB.Eband.max(),NE)
    
    DOS = np.zeros(len(Elin))
    for ki in range(len(tetra)):
        sys.stdout.write('\r'+progress_bar(ki+1,len(tetra)))
        for bi in range(len(TB.basis)): #iterate over all bands
            DOS += band_contribution(TB.Eband[tetra[ki]][:,bi],Elin,len(tetra))
    print('DOS calculation complete')
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.plot(Elin,DOS)               
    return Elin,DOS

def band_contribution(eigenvals,w_domain,volume):
    
    '''

    Compute the contribution over a single tetrahedron, from a 
    single band, to the density of states
    
    *args*:

        - **eigenvals**: numpy array of float, energy values at corners
        
        - **w_domain**: numpy array of float, energy domain
        
        - **volume**: int, number of tetrahedra in the total mesh
    
    *return*:

        - **DOS**: numpy array of float, same length as w_domain
    
    ***
    '''
    eig_sort = sorted(eigenvals)
    args = (*eig_sort,1,volume)
    DOS = dos_func(w_domain,args)
    
    return DOS

########################Partial Density of States##############################

    
def proj_avg(eivecs,proj_matrix):
    '''
    Calculate the expectation value of the projection operator, for each of the
    eigenvectors, at each of the vertices, and then sum over the vertices. We
    use *numpy.einsum* to perform matrix multiplication and contraction.
    
    *args*:

        - **eivecs**: numpy array of complex float, 4xNxM, with M number of eigenvectors,
        N basis dimension
        
        - **proj_matrix**: numpy array of complex float, NxN in size
        
    *return*:

        - numpy array of M float, indicating the average projection over the 4 
        corners of the tetrahedron
      
    ***
    '''
    return np.real(0.25*np.einsum('ijk,ijk->k',np.conj(eivecs),np.einsum('ij,kjl->kil',proj_matrix,eivecs)))

def proj_mat(proj,lenbasis):
    '''
    Define projection matrix for fast evaluation of the partial density of states
    weighting. As the projector here is diagonal, and represents a Hermitian 
    matrix, it is by definition a real matrix operator.
    
    *args*:

        - **proj**: numpy array, either 1D (indices of projection), or 2D (indices of
        projection and weight of projection)
        
        - **lenbasis**: int, size of the orbital basis
    
    *return*:

        - **projector**: numpy array of float, lenbasis x lenbasis
    
    ***
    '''
    projector = np.identity(lenbasis,dtype=complex)
    
    proj_vect = np.zeros(lenbasis,dtype=complex)

    if len(np.shape(proj))==1:
        proj_vect[proj] = 1/(len(proj))**0.5
    
    elif len(np.shape(proj))==2:
        proj_vect[proj[:,0]] = proj[:,1]
        proj_vect/=np.sqrt(np.einsum('i,i',np.conj(proj_vect),proj_vect))
    
    projector*=np.real(proj_vect)
    
    return projector
        
        
    


def pdos_tetra(TB,NE,NK,proj):
    
    '''
    Partial density of states calculation. Follows same tetrahedra method, 
    weighting the contribution of a given tetrahedra by the average projection
    onto the indicated user-defined projection. The average here taken as the sum
    over projection at the 4 vertices of the tetrahedra.
    
    *args*:

        - **TB**: tight-binding model object
        
        - **NE**: int, number of energy bins
        
        - **NK**: int, or iterable of 3 int, indicating the number of k-points
        along each of the axes of the Brillouin zone
        
        - **proj**: numpy array of float, 1D or 2D, c.f. *proj_mat*.
        
    *return*:

        - **Elin**: numpy array of float, with length **NE**, spanning the
        range of the tight-binding bandstructure
        
        - **pDOS**: numpy array of float, len **NE**, projected density of states
        
        - **DOS**: numpy array of float, len **NE**, full density of states
    
    ***
    '''
    
    projection_matrix = proj_mat(proj,len(TB.basis))
    
    kpts,tetra = tetrahedra.mesh_tetra(TB.avec,NK)
    print('k tetrahedra defined')
    TB.Kobj.kpts = kpts
    TB.solve_H()
    print('Diagonalization complete')
    Elin = np.linspace(TB.Eband.min(),TB.Eband.max(),NE)
    
    DOS = np.zeros(len(Elin))
    pDOS = np.zeros(len(Elin))
    for ki in range(len(tetra)):
        eivecs = TB.Evec[tetra[ki],:,:]
        projection_avg = proj_avg(eivecs,projection_matrix)
        sys.stdout.write('\r'+progress_bar(ki+1,len(tetra)))
        for bi in range(len(TB.basis)): #iterate over all bands
            DOS_tetra = band_contribution(TB.Eband[tetra[ki]][:,bi],Elin,len(tetra))
            pDOS += DOS_tetra*projection_avg[bi]
            DOS += DOS_tetra
    
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.plot(Elin,DOS)
    ax.plot(Elin,pDOS)
    
    return Elin,pDOS,DOS
    
   
##############################-------D(E)---------#############################
def dos_func(energy,epars):
    '''

    Piecewise function for calculation of density of states
    
    *args*:

        - **energy**: numpy array of float (energy domain)
        
        - **epars**: tuple of parameters: e[0],e[1],e[2],e[3],V_T,V_G being the ranked band energies for the tetrahedron, 
        as well as the volume of both the tetrahedron and the Brillouin zone, all float
    
    *return*:

        - numpy array of float giving DOS contribution from this tetrahedron
    
    ***
    '''
    print(epars)
    return np.piecewise(energy,[energy<epars[0],(epars[0]<=energy)*(energy<epars[1]),(epars[1]<=energy)*(energy<epars[2]),(epars[2]<=energy)*(energy<epars[3]),energy>=epars[3]],[e_out,e_12,e_23,e_34,e_out],epars)


def e_out(energy,epars):
    return np.zeros(len(energy))

def e_12(energy,epars):
    return epars[4]/epars[5]*3*(energy-epars[0])**2/(epars[1]-epars[0])/(epars[2]-epars[0])/(epars[3]-epars[0])

def e_23(energy,epars):
    e21,e31,e41,e42,e32 = epars[1]-epars[0],epars[2]-epars[0],epars[3]-epars[0],epars[3]-epars[1],epars[2]-epars[1]
    e2 = energy-epars[1]
    return epars[4]/epars[5]/e31/e41*(3*e21+6*e2-3*(e31+e42)/(e32*e42)*e2**2)

def e_34(energy,epars):
    return epars[4]/epars[5]*3*(epars[3]-energy)**2/(epars[3]-epars[0])/(epars[3]-epars[1])/(epars[3]-epars[2])
##############################-------D(E)---------#############################
    

##############################-------n(E)---------#############################
def find_EF_tetra_dos(TB,occ,dE,NK):

    '''
    Use the tetrahedron-integration method to establish the Fermi-level, for a given
    electron occupation.
    
    *args*:

        - **TB**: instance of tight-binding model object from *TB_lib*

        - **occ**: float, desired electronic occupation
        
        - **dE**: estimate of energy precision desired for evaluation of the 
        Fermi-level (in eV)
        
        - **NK**: int or iterable of 3 int, number of k points in mesh.
        
    *return*:

        **EF**: float, Fermi Energy for the desired occupation, to within dE of actual
        value.
    
    ***
    '''
    e_domain,n_elec = n_tetra(TB,dE,NK)
    EF = e_domain[np.where(abs(n_elec-occ)==abs(n_elec-occ).min())[0][0]]
    return EF


def n_tetra(TB,dE,NK,plot=True):
    '''
    This function, also from the algorithm of Blochl, gives the integrated DOS
    at every given energy (so from bottom of bandstructure up to its top. This makes
    for very convenient and precise evaluation of the Fermi level, given an electron
    number)
    
    *args*:

        - **TB**: tight-binding model object
        
        - **dE**: float, energy spacing (meV)
        
        - **NK**: int, iterable of 3 int. number of k-points in mesh
        
        - **plot**: bool, to plot or not to plot the calculated array
    
    *return*:

        - **Elin**: linear energy array of float, spanning the range of the eigenspectrum
        
        - **n_elect**: numpy array of float, same length as **Elin**, integrated DOS 
        at each energy, i.e. total number of electrons occupied at each energy
        
    ***
    '''
    kpts,tetra = tetrahedra.mesh_tetra(TB.avec,NK)
    TB.Kobj.kpts = kpts
    TB.solve_H()
    Elin = np.arange(TB.Eband.min(),TB.Eband.max(),dE)
    n_elect = np.zeros(len(Elin))
    for ki in range(len(tetra)):
        E_tmp = TB.Eband[tetra[ki]]
        for bi in range(len(TB.basis)): #iterate over all bands
            Eband = sorted(E_tmp[:,bi])
            args = (*Eband,1,len(tetra))
            n_elect += n_func(Elin,args)
    if plot:
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(Elin,n_elect)               
    return Elin,n_elect    



def n_func(energy,epars):

    '''
    Piecewise function for evaluating contribution of tetrahedra to electronic
    occupation number
    
    *args*:

        - **energy**: numpy array of float, energy domain
        
         - **epars**: tuple of parameters: e[0],e[1],e[2],e[3],V_T,V_G being the ranked band energies for the tetrahedron, 
        as well as the volume of both the tetrahedron and the Brillouin zone, all float
       
    *return*:
    
        - numpy array of float, same length as **energy**, providing contribution of
        tetrahedra to the occupation function
        
    ***
    '''
    
    
    return np.piecewise(energy,[energy<epars[0],(epars[0]<=energy)*(energy<epars[1]),(epars[1]<=energy)*(energy<epars[2]),(epars[2]<=energy)*(energy<epars[3]),energy>=epars[3]],[n1,n12,n23,n34,n4],epars)

def n1(energy,epars):
    return np.zeros(len(energy))

def n12(energy,epars):
    return epars[4]/epars[5]*(energy-epars[0])**3/(epars[1]-epars[0])/(epars[2]-epars[0])/(epars[3]-epars[0])

def n23(energy,epars):
    e21,e31,e41,e42,e32 = epars[1]-epars[0],epars[2]-epars[0],epars[3]-epars[0],epars[3]-epars[1],epars[2]-epars[1]
    e2 = energy-epars[1]
    return epars[4]/epars[5]*(1/(e31*e41))*(e21**2+3*e21*(e2)+3*e2**2-(e31+e42)/(e32*e42)*(e2**3))

def n34(energy,epars):
    return epars[4]/epars[5]*(1-(epars[3]-energy)**3/(epars[3]-epars[0])/(epars[3]-epars[1])/(epars[3]-epars[2]))

def n4(energy,epars):
    return epars[4]/epars[5]

##############################-------n(E)---------#############################
    
    
def progress_bar(N,Nmax):
    frac = N/Nmax
    st = ''.join(['|' for i in range(int(frac*30))])
    st = '{:30s}'.format(st)+'{:3d}%'.format(int(frac*100))
    return st
    
    




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#Created on Mon Nov 13 15:41:01 2017
#@author: rday

#MIT License

#Copyright (c) 2018 Ryan Patrick Day

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.
import numpy as np
import matplotlib.pyplot as plt
from operator import itemgetter
from itertools import compress
import sys

if sys.version_info<(3,0):
    print('Warning: This software requires Python 3.0 or higher. Please update your Python instance before proceeding')
else:
    import chinook.H_library as Hlib
    
try:
    import psutil
    ps_found = True
except ModuleNotFoundError:
    print('psutil not found, please load for better memory handling. See documentation for more detail')
    ps_found = False


'''
Tight-Binding Utility module

'''


class H_me:
    '''
    This class contains the relevant executables and data structure pertaining 
    to generation of the Hamiltonian matrix elements for a single set of 
    coupled basis orbitals. Its attributes include integer values 
    **i**, **j** indicating the basis indices, and a list of hopping
    vectors/matrix element values for the Hamiltonian. 
    
    The method **H2Hk** provides an executable function of momentum to allow
    broadcasting of the Hamiltonian over a large array of momenta.
    Python's flexible protocol for equivalency and passing variables by
    reference/value require definition of a copy operator which allows one to
    produce safely, a copy of the object rather than its coordinates 
    in memory alone.
    ***
    '''
    def __init__(self,i,j,executable=False):
        '''
        Initialize the H_me object, the related basis indices and an empty list
        of hopping elements.
        
        *args*:

            - **i**, **j**: int, basis indices

        *kwargs*:

            - **executable**: boolean, True if using unconventional, executable form of Hamiltonian rather
            than standard tij type tight-binding model
            
        ***
        '''
        self.executable = executable
        self.i = i
        self.j = j
        self.H = []
    
    def append_H(self,H,R0=0,R1=0,R2=0):
        '''
        Add a new hopping path to the coupling of the parent orbitals.
        
        *args*:

            - **H**: complex float, matrix element strength, or if self.exectype,
            should be an executable 
            
            - **R0**, **R1**, **R2**: float connecting vector in cartesian
            coordinate frame--this is the TOTAL vector, not the relevant 
            lattice vectors only
            
            
        *return*:

            - directly modifies the Hamiltonian list for these matrix
            coordinates

        ***
        '''
        if not self.executable:
            self.H.append([R0,R1,R2,H])
        else:
            self.H.append(H)
        
    def H2Hk(self):  
        '''
        Transform the list of hopping elements into a Fourier-series expansion 
        of the Hamiltonian. This is run during diagonalization for each
        matrix element index. If running a low-energy Hamiltonian, executable functions are
        simply summed for each basis index i,j, rather than computing a Fourier series. x is
        implicitly a numpy array of Nx3: it is essential that the executable conform to this input type.
        
        *return*:

            - lambda function of a numpy array of float of length 3
            
        ***
        '''
        if not self.executable:
            return lambda x: sum([complex(m[-1])*np.exp(1.0j*np.dot(x,np.array([m[0],m[1],m[2]]))) for m in self.H])
        return lambda x: sum([m(x) for m in self.H])
        
    def clean_H(self):
        
        '''
        
        Remove all duplicate instances of hopping elements in the matrix 
        element list. This function is run automatically during slab generation.
        
        The Hamiltonian list is not itself directly modified. 
        
        *return*:

            - list of hopping vectors and associated Hamiltonian matrix
            element strengths
            
        ***
        '''
        Hel_double = self.H
        bools = [True]*len(Hel_double)
        for hi in range(len(Hel_double)-1):
            for hj in range(hi+1,len(Hel_double)):
                
                try:
                    norm = np.linalg.norm(np.array(Hel_double[hi])-np.array(Hel_double[hj]))
                    if abs(norm)<1e-5:
                        bools[hj] = False
                except IndexError:
                    continue
        return list(compress(Hel_double,bools))
    
    def copy(self):
        
        '''
        Copy by value of the **H_me** object

        *return*:

            - **H_copy**: duplicate **H_me** object
       
        ***
        '''
        
        H_copy = H_me(self.i,self.j)
        H_copy.H = [h for h in self.H]
        return H_copy
    
    
        
class TB_model:
    '''
    The **TB_model** object carries the model basis as a list of **orbital**
    objects, as well as the model Hamiltonian, as a list of **H_me**.
    '''
    
    def __init__(self,basis,H_args,Kobj = None):
        '''
        Initialize the tight-binding model object.
        
        *args*:  

            - **basis**: list of orbital objects
                    
            - **H_args**: dictionary for Hamiltonian build:
            
                - *'type'*: string,  "SK" ,"list", or "txt"
             
            - if *'type'* is *'SK'*:
                - *'V'*: dictionary of Slater-Koster hopping terms, 
                c.f. **chinook.H_library**
                
                - *'avec'*: numpy array of 3x3 float,
                lattice vectors for generating neighbours
                 
                
             - if *'type'* is *'list'*:
                 
                 - *'list'*: list of Hamiltonian elements: [i,j,R1,R2,R3,Hij]
                 
             - if *'type'* is 'txt':
                 
                 - *'filename'*:path to text file
                 
             - *'cutoff'*: float, cutoff distance for hopping
             
             - *'renorm'*: float, renormalization factor
                 
             - *'offset'*: float, offset energy, in eV
             
             - *'tol'*: float, minimum amplitude for matrix element term
            
             - *'spin'* : dictionary, with key values of 'bool':boolean,
             'soc':boolean, 'lam': dictionary of integer:float pairs
        
        *kwargs*:

            - **Kobj**: momentum object, as defined in *chinook.klib.py*
            
        ***
        '''
        self.basis = basis 
    
        if H_args is not None:
            if 'avec' in H_args.keys():
                self.avec = H_args['avec']
        
            self.mat_els = self.build_ham(H_args)
            
        self.Kobj = Kobj

        
    def copy(self):
        '''
        Copy by value of the **TB_model** object
        
        *return*:

            - **TB_copy**: duplicate of the **TB_model** object.
        '''
        
        basis_copy = [o.copy() for o in self.basis]
        TB_copy = TB_model(basis_copy,None,self.Kobj)
        TB_copy.avec = self.avec.copy()
        TB_copy.mat_els = [m.copy() for m in self.mat_els]
        
        return TB_copy
    
    def print_basis_summary(self):
        '''
        Very basic print function for printing a summary
        of the orbital basis, including their label, atomic species, position
        and spin character.
        '''
        print(' Index | Atom | Label | Spin |     Position     ')
        print('================================================')
        for o in self.basis:
            print('  {:3d}  |  {:2d}  |{:7}|{:6}| {:0.03f},{:0.03f},{:0.03f}'.format(o.index,o.atom,o.label,0.5*o.spin,*o.pos))
        
            
    
    def build_ham(self,H_args):
        
        '''
        Buld the Hamiltonian using functions from **chinook.H_library.py**

        *args*:

            - **H_args**: dictionary, containing all relevant information for
            defining the Hamiltonian list. For details, see **TB_model.__init__**.
            
        *return*:

            - sorted list of matrix element objects. These objects have 
            i,j attributes referencing the orbital basis indices, 
            and a list of form [R0,R1,R2,Re(H)+1.0jIm(H)]
        
        ***
        '''
        executable = False
        if type(H_args)==dict:
            try:
                ham_list = []
                if H_args['type'] == "SK":
                    ham_list = Hlib.sk_build(H_args['avec'],self.basis,H_args['V'],H_args['cutoff'],H_args['tol'],H_args['renorm'],H_args['offset'])
                elif H_args['type'] == "txt":
                    ham_list = Hlib.txt_build(H_args['filename'],H_args['cutoff'],H_args['renorm'],H_args['offset'],H_args['tol'])
                elif H_args['type'] == "list":
                    ham_list = H_args['list']
                elif H_args['type'] == 'exec':
                    ham_list = H_args['exec']
                    executable = True
                if 'spin' in H_args.keys():
                    if H_args['spin']['bool']:
                        ham_spin_double = Hlib.spin_double(ham_list,len(self.basis))
                        ham_list = ham_list + ham_spin_double   
                        if H_args['spin']['soc']:
                            ham_so = Hlib.SO(self.basis)
                            ham_list = ham_list + ham_so
                        if 'order' in H_args['spin']:
                            if H_args['spin']['order']=='F':
                                ham_FM = Hlib.FM_order(self.basis,H_args['spin']['dS'])
                                ham_list = ham_list + ham_FM
                            elif H_args['spin']['order']=='A':
                                ham_AF = Hlib.AFM_order(self.basis,H_args['spin']['dS'],H_args['spin']['p_up'],H_args['spin']['p_dn'])
                                ham_list = ham_list + ham_AF
                H_obj = gen_H_obj(ham_list,executable)
                return H_obj
            except KeyError:
                print('Invalid dictionary input for Hamiltonian generation.')
                return None
        else:
            return None
        
    def append_H(self,new_elements):
        '''
        Add new terms to the Hamiltonian by hand. This directly modifies
        the list of Hamiltonian matrix element, self.mat_els of the TB object.
        
        *args*:

            - **new_elements**: list of Hamiltonian matrix elements, either a single element [i,j,x_ij,y_ij,z_ij,H_ij(x,y,z)]
            or as a list of such lists. Here i, j are the related orbital-indices.
        
        ***
        '''
        if type(new_elements[0])!=list:
            new_elements = [new_elements]
        for elmts in new_elements:
            i,j = elmts[:2]
            new_pair = True
            for Hme in self.mat_els:
                if Hme.i==i and Hme.j==j:
                    Hme.H.append(elmts[2:])
                    new_pair = False
            if new_pair:
                self.mat_els.append(H_me(i,j))
                self.mat_els[-1].append_H(*elmts[2:])
        
    def unpack(self):
        '''
        Reduce a Hamiltonian object down to a list of matrix elements. Include the Hermitian conjugate terms
    
    
        *return*:

            - **Hlist**: list of Hamiltonian matrix elements
            
        ***
        '''
        Hlist =[]
        for hij in self.mat_els:
            for el in hij.H:
                Hlist.append([hij.i,hij.j,*el])
        return Hlist
        
        
    def solve_H(self,Eonly = False):
        '''
        This function diagonalizes the Hamiltonian over an array of momentum vectors.
        It uses the **mat_el** objects to quickly define lambda functions of 
        momentum, which are then filled into the array and diagonalized.
        According to https://ntrs.nasa.gov/archive/nasa/casi.ntrs.nasa.gov/20050192421.pdf
        SVD algorithms require memory of 2*order*(4*order + 1) ~ 8*order^2. The matrices are
        complex float, so this should be 16 bytes per entry: so len(k)*(2048*order**2). If 
        the diagonalization is requesting more than 85% of the available memory, then split
        up the k-path into sequential diagonalizations.
        
        *return*:

            - **self.Eband**: numpy array of float, shape(len(self.Kobj.kpts),len(self.basis)),
            eigenvalues
            
            - **self.Evec**: numpy array of complex float, shape(len(self.Kobj.kpts),len(self.basis),len(self.basis))
            eigenvectors
                
        ***
        '''
        partition = False
        if ps_found:
            mem_summary = psutil.virtual_memory()
            avail = mem_summary.available
            size_lim = int(0.85*avail)
            mem_req = int(len(self.Kobj.kpts)*len(self.basis)**2*2048)
            if mem_req>size_lim:
                partition = True
                N_partitions = int(np.ceil(mem_req/size_lim))
                splits = [j*int(np.floor(len(self.Kobj.kpts)/N_partitions)) for j in range(N_partitions)]
                splits.append(len(self.Kobj.kpts))
                print('Large memory load: splitting diagonalization into {:d} segments'.format(N_partitions))
 
        if self.Kobj is not None:
            Hmat = np.zeros((len(self.Kobj.kpts),len(self.basis),len(self.basis)),dtype=complex) #initialize the Hamiltonian
            
            for me in self.mat_els:
                Hfunc = me.H2Hk() #transform the array above into a function of k
                Hmat[:,me.i,me.j] = Hfunc(self.Kobj.kpts) #populate the Hij for all k points defined
            if not partition:
                if not Eonly:
                    self.Eband,self.Evec = np.linalg.eigh(Hmat,UPLO='U') #diagonalize--my H_raw definition uses i<=j, so we want to use the upper triangle in diagonalizing
                else:
                    self.Eband = np.linalg.eigvalsh(Hmat,UPLO='U') #get eigenvalues only
                    self.Evec = np.array([0])
            else:
                self.Eband = np.zeros((len(self.Kobj.kpts),len(self.basis)))
                self.Evec =  np.zeros((len(self.Kobj.kpts),len(self.basis),len(self.basis)),dtype=complex)
                for ni in range(len(splits)-1):
                    self.Eband[splits[ni]:splits[ni+1]],self.Evec[splits[ni]:splits[ni+1]] = np.linalg.eigh(Hmat[splits[ni]:splits[ni+1]],UPLO ='U')
            return self.Eband,self.Evec
        else:
            print('You have not defined a set of kpoints over which to diagonalize.')
            return False
            
        
        
    def plotting(self,win_min=None,win_max=None,ax=None): #plots the band structure. Takes in Latex-format labels for the symmetry points indicated in the main code
        '''
        Plotting routine for a tight-binding model evaluated over some path in k.
        If the model has not yet been diagonalized, it is done automatically
        before proceeding.
        
        *kwargs*:

            - **win_min**, **win_max**: float, vertical axis limits for plotting
            in units of eV. If not passed, a reasonable choice is made which 
            covers the entire eigenspectrum.
        
            - **ax**: matplotlib Axes, for plotting on existing Axes
            
        *return*:

            - **ax**: matplotlib axes object
        
        ***
        '''
        try:
            Emin,Emax = np.amin(self.Eband),np.amax(self.Eband)
        except AttributeError:
            print('Bandstructure and energies have not yet been defined. Diagonalizing now.')
            self.solve_H()
            Emin,Emax = np.amin(self.Eband),np.amax(self.Eband)
        
        if ax is None:
            fig=plt.figure()
            fig.set_tight_layout(False)
            ax=fig.add_subplot(111)
            
        ax.axhline(y=0,color='k',lw=1.5,ls='--')
        for b in self.Kobj.kcut_brk:
            ax.axvline(x = b,color = 'k',ls='--',lw=1.5)
        for i in range(len(self.basis)):
            ax.plot(self.Kobj.kcut,np.transpose(self.Eband)[i,:],color='navy',lw=1.5)

        plt.xticks(self.Kobj.kcut_brk,self.Kobj.labels)
        if win_max==None or win_min==None:
            ax.set_xlim(self.Kobj.kcut[0],self.Kobj.kcut[-1])
            ax.set_ylim(Emin-1.0,Emax+1.0)
        elif win_max !=None and win_min !=None:
            ax.set_xlim(self.Kobj.kcut[0],self.Kobj.kcut[-1])
            ax.set_ylim(win_min,win_max) 
        ax.set_ylabel("Energy (eV)")

        return ax  
    
    def plot_unitcell(self,ax=None):
        
        '''
        Utility script for visualizing the lattice and orbital basis.
        Distinct atoms are drawn in different colours
        
        *kwargs*:

            - **ax**: matplotlib Axes, for plotting on existing Axes
            
        *return*:

            - **ax**: matplotlib Axes, for further modifications to plot
        
        ***
        '''
        edges = cell_edges(self.avec)
        coord_dict = atom_coords(self.basis)

        
        fig = plt.figure()
        ax = fig.add_subplot(111,projection='3d')
        for ed in edges:
            ax.plot(ed[:,0],ed[:,1],ed[:,2],c='k')
        for atoms in coord_dict.keys():
            ax.scatter(coord_dict[atoms][:,0],coord_dict[atoms][:,1],coord_dict[atoms][:,2],s=20)
            
        return ax
          
            
def gen_H_obj(htmp,executable=False):
    
    '''
    Take a list of Hamiltonian matrix elements in list format:
    [i,j,Rij[0],Rij[1],Rij[2],Hij(R)] and generate a list of **H_me**
    objects instead. This collects all related matrix elements for a given
    orbital-pair for convenient generation of the matrix Hamiltonians over
    an input array of momentum
    
    *args*:

        - **htmp**: list of numeric-type values (mixed integer[:2], float[2:5], complex-float[-1])
    
    *kwargs*:

        - **executable**: boolean, if True, we don't have a standard Fourier-type Hamiltonian,
        but perhaps a low-energy expansion. In this case, the htmp elements are
        
    *return*:

        - **Hlist**: list of Hamiltonian matrix element, **H_me** objects
    
    ***
    '''
    if not executable:
        htmp = sorted(htmp,key=itemgetter(0,1,2,3,4))
    else:
        htmp = sorted(htmp,key=itemgetter(0,1))
    
    Hlist = []
    Hnow = H_me(int(htmp[0][0]),int(htmp[0][1]),executable=executable)
    Rij = np.zeros(3)
    
    for h in htmp:
        if h[0]!=Hnow.i or h[1]!=Hnow.j:
            Hlist.append(Hnow)
            Hnow = H_me(int(np.real(h[0])),int(np.real(h[1])),executable=executable)
        if not executable:
            Rij = np.real(h[2:5])
            Hnow.append_H(H=h[5],R0=Rij[0],R1=Rij[1],R2=Rij[2])
        else:
            Hnow.append_H(H=h[2])
    Hlist.append(Hnow)
    return Hlist 
            
        
def cell_edges(avec):

    '''
    Define set of line segments which enclose the unit cell. 
    
    *args*:

        - **avec**: numpy array of 3x3 float
        
    *return*:

        - **edges**: numpy array of 12 x 6, endpoints of the 12 edges of the unit cell parallelepiped
    
    ***
    '''
    
    modvec = np.array([[np.mod(int(j/4),2),np.mod(int(j/2),2),np.mod(j,2)] for j in range(8)])
    edges = []
    for p1 in range(len(modvec)):
        for p2 in range(p1,len(modvec)):
            if np.linalg.norm(modvec[p1]-modvec[p2])==1:
                edges.append([np.dot(avec.T,modvec[p1]),np.dot(avec.T,modvec[p2])])
    edges = np.array(edges)
    return edges


def atom_coords(basis):
    '''
    Define a dictionary organizing the distinct coordinates of instances of each
    atomic species in the basis
    
    *args*:

        - **basis**: list of orbital objects
        
    *return*:

        - **dictionary with integer keys, numpy array of float values. atom:locations are
        encoded in this way

    ***.
    '''
    
    coord_dict = {}
    all_pos = [[o.atom,*o.pos] for o in basis]
    for posns in all_pos:
        if posns[0] not in coord_dict.keys():
            coord_dict[posns[0]] = [np.array(posns[1:])]
        else:

            min_dist = np.array([np.sqrt((posns[1]-c[0])**2+(posns[2]-c[1])**2+(posns[3]-c[2])**2) for c in coord_dict[posns[0]]]).min()
            if min_dist>0:
                coord_dict[posns[0]].append(np.array(posns[1:]))
    for atoms in coord_dict:
        coord_dict[atoms] = np.array(coord_dict[atoms])
    return coord_dict
            
       
    
    
        
    
    



# -*- coding: utf-8 -*-

#Created on Sun Jul 15 13:49:17 2018
#@author: ryanday
#MIT License

#Copyright (c) 2018 Ryan Patrick Day

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

import numpy as np

def ang_v1v2(v1,v2):
    '''
    Find angle between two vectors:
        
    *args*:

        - **v1**: numpy array of 3 float
        
        - **v2**: numpy array of 3 float
        
    *return*:

        - float, angle between the vectors
        
    ***
    '''
    
    return np.arccos(np.around(np.dot(v1,v2)/(np.linalg.norm(v1)*np.linalg.norm(v2)),15))

def are_parallel(v1,v2):
    
    '''
    Determine if two vectors are parallel:
        
    *args*:

        - **v1**: numpy array of 3 float
        
        - **v2**: numpy array of 3 float
        
    *return*:

        - boolean, True if parallel to within 1e-5 radians
   
    ***
    '''
    
    cos = np.dot(v1,v2)/(np.linalg.norm(v1)*np.linalg.norm(v2))
    if abs(cos)>1e-5 and np.mod(cos,1)<1e-5:
        return True
    else:
        return False

def are_same(v1,v2):
    
    '''
    Determine if two vectors are identical
    
    *args*:

        - **v1**: numpy array of 3 float
        
        - **v2**: numpy array of 3 float
        
    *return*:

        - boolean, True if the two vectors are parallel and have same
        length, both to within 1e-5

    ***
    '''
    
    if are_parallel(v1,v2):
        mod_val = (np.mod(np.dot(v1,v2)/np.linalg.norm(v2)**2,1))
        if mod_val<1.0e-5 or mod_val>0.99999:
            return True
        else:
            return False
    else:
        return False
    
    
def initialize_search(v1,v2,avec):
    
    '''
    Seed search for v3 with the nearest-neighbouring Bravais latice point
    which maximizes the projection out of plane of that spanned by v1 and v2.
    
    *args*: 

        - **v1**, **v2**: numpy array of 3 float, the spanning vectors for plane 
    
        - **avec**: numpy array of 3x3 float
        
    *return*:

        - numpy array of 3 float, the nearby Bravais lattice point which
        maximizes the projection along the plane normal
        
    ***
    '''
    nv = np.cross(v1,v2)
    near = np.array([np.dot([int(np.mod(i,(3)))-1,int(np.mod(i/(3),(3)))-1,int(i/(3)**2)-1],avec) for i in range((3)**3) if i!=int((3)**3/2)])
    angs = np.array([ang_v1v2(ni,nv) for ni in near])
    choice = np.where(abs(angs)==abs(angs).min())[0]
    return near[choice[0]]


def refine_search(v3i,v1,v2,avec,maxlen):
    '''
    Refine the search for the optimal v3--supercell lattice vector which both 
    minimizes its length, while maximizing orthogonality with v1 and v2
    
    *args*:

        - **v3i**: numpy array of 3 float, initial guess for v3
        
        - **v1**: numpy array of 3 float, in-plane supercell vector
        
        - **v2**: numpy array of 3 float, in-plane supercell vector
        
        - **avec**: numpy array of 3x3 float, bulk lattice vectors
        
        - **maxlen**: float, upper limit on how long of a third vector we can
        reasonably tolerate. This becomes relevant for unusual Miller indices.
        
    *return*:

        - **v3_opt** list of numpy array of 3 float, list of viable options for 
        the out of plane surface unit cell vector
        
    ***
    '''
    nv = np.cross(v1,v2)
    atol = 1e-2
    v_add = np.array([-v2,-v1,np.zeros(3),v1,v2])
    v3_opt = []
    for qi in range(1,50):
        tmp_v = qi*v3i
        ang_to_norm = ang_v1v2(tmp_v,nv)
        if abs(ang_to_norm)<atol:
            ok = True
        else:
            ok = False
        counter = 1
        while not ok:
            v_opt = tmp_v + counter*v_add
            angles = np.array([ang_v1v2(vi,nv) for vi in v_opt])
            try:
                better = abs(angles[np.where(abs(angles)<abs(ang_to_norm))[0]]).min()
                tmp_v = v_opt[np.where(abs(angles)==better)[0][0]]
                ang_to_norm = ang_v1v2(tmp_v,nv)
                if abs(ang_v1v2(tmp_v,nv))<atol:
                    ok = True

            except ValueError:
                ok = True
            counter+=1
        if np.linalg.norm(tmp_v)<maxlen:
            v3_opt.append(tmp_v)
    return v3_opt

def score(vlist,v1,v2,avec):
    
    '''
    To select the ideal out-of-plane surface unit cell vector, score the 
    candidates based on both their length and their orthogonality with respect
    to the two in-plane spanning vectors. The lowest scoring candidate is selected
    as the ideal choice.
    
    *args*:

        - **vlist**: list of len 3 numpy array of float, choices for out-of-plane
        vector
        
        - **v1**, **v2**: numpy array of 3 float, in plane spanning vectors
        
        - **avec**: numpy array of 3x3 float, primitive unit cell vectors
        
    *return*:

        - numpy array of len 3, out of plane surface-projected lattice vector
   
    ***
    '''
    nv = np.cross(v1,v2)
    
    angles = np.array([ang_v1v2(vi,nv) for vi in vlist])
    if abs(angles).max()>0.0:
        angles/=abs(angles).max()
    len_proj = np.array([np.linalg.norm(vi) for vi in vlist])
    if abs(len_proj).max()>0.0:
        len_proj/=abs(len_proj).max()
    score_vec = np.array([[len_proj[i],angles[i]] for i in range(len(vlist))])
    try:
        return vlist[np.where(np.linalg.norm(score_vec)==np.linalg.norm(score_vec).min())[0][0]]
    except IndexError:
        print('Require larger search field, increasing range of acceptance for surface lattice vectors.')
        return np.zeros(3)               

def find_v3(v1,v2,avec,maxlen):
    '''
    Find the best out-of-plane surface unit cell vector. While we initialize with
    a fixed cutoff for maximum length, to avoid endless searching, we can slowly 
    increase on each iteration until a good choice is possible.
    
    *args*:

        - **v1**, **v2**: numpy array of 3 float, in plane spanning vectors
        
        - **avec**: numpy array of 3x3 float, bulk lattice vectors
        
        - **maxlen**: float, max length tolerated for the vector we seek
        
    *return*:
    
        - **v3_choice**: the chosen unit cell vector
        
    ***
    '''
    
    v3i = initialize_search(v1,v2,avec)
    found = False
    while not found:
        v3f = refine_search(v3i,v1,v2,avec,maxlen)
        v3_choice = score(v3f,v1,v2,avec)
        if np.linalg.norm(v3_choice)>0:
            found = True
        else:
            maxlen *=1.2

    return v3_choice
                    
            
    
    
    


#if __name__ == "__main__":
#    avec = np.array([[0,4.736235,0],[4.101698,-2.368118,0],[0,0,13.492372]])
#    miller = np.array([1,0,4])
    
#    a =  3.52
#    avec = np.array([[a/2,a/2,0],[0,a/2,a/2],[a/2,0,a/2]])
#    miller = np.array([1,1,1])
#    v12 = v_vecs(miller,avec)
#    maxlen = 40
#    v3 = find_v3(v12[0],v12[1],avec,maxlen)